# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import csv
import importlib.util
import logging
import os
import re
import warnings
from datetime import datetime, timedelta

import numpy as np
import tifffile
from dateutil import parser
from tifffile import TiffFile, TiffPage, imwrite

from rsciio._docstrings import (
    FILENAME_DOC,
    LAZY_DOC,
    RETURNS_DOC,
    SIGNAL_DOC,
)
from rsciio.utils._date_time import get_date_time_from_metadata
from rsciio.utils._dictionary import DTBox

_logger = logging.getLogger(__name__)


axes_label_codes = {
    "X": "width",
    "Y": "height",
    "S": "sample",
    "P": "plane",
    "I": "image series",
    "Z": "depth",
    "C": "color|em-wavelength|channel",
    "E": "ex-wavelength|lambda",
    "T": "time",
    "R": "region|tile",
    "A": "angle",
    "F": "phase",
    "H": "lifetime",
    "L": "exposure",
    "V": "event",
    "Q": None,
    "_": None,
}


def file_writer(filename, signal, export_scale=True, extratags=None, **kwds):
    """
    Write data to tif using Christoph Gohlke's ``tifffile`` library.

    Parameters
    ----------
    %s
    %s
    export_scale : bool, default=True
        Export the scale and the units (compatible with DM and ImageJ) to
        appropriate tags.
    extratags : tuple, list of tuple, None, default=None
        Save custom tags through the ``tifffile`` library. Must conform to a
        specific convention (see `tifffile documentation
        <https://github.com/cgohlke/tifffile>`_ and example below).
    **kwds : dict, optional
        Additional arguments to be passed to the ``imwrite`` function of the `tifffile library
        <https://github.com/cgohlke/tifffile>`_.

    Examples
    --------
    >>> # Saving the string 'Random metadata' in a custom tag (ID 65000)
    >>> extratag = [(65000, 's', 1, "Random metadata", False)]
    >>> file_writer('file.tif', signal, extratags=extratag)

    >>> # Reading the string 'Random metadata' from a custom tag (ID 65000)
    >>> s2 = file_reader('file.tif')
    >>> s2.original_metadata['Number_65000']
    'Random metadata'
    """

    data = signal["data"]
    metadata = signal.get("metadata", {})
    photometric = "MINISBLACK"
    # HyperSpy uses struct arrays to store RGBA data
    from rsciio.utils import rgb

    if extratags is None:
        extratags = []

    if rgb.is_rgbx(data):
        data = rgb.rgbx2regular_array(data)
        photometric = "RGB"

    if "description" in kwds.keys() and export_scale:
        kwds.pop("description")
        _logger.warning(
            "Description and export scale cannot be used at the same time, "
            "because it is incompability with the 'ImageJ' tiff format"
        )
    if export_scale and "axes" in signal.keys():
        kwds.update(_get_tags_dict(signal, extratags=extratags))
        _logger.debug(f"kwargs passed to tifffile.py imsave: {kwds}")

        if "metadata" not in kwds.keys():
            # Because we write the calibration to the ImageDescription tag
            # for imageJ, we need to disable tiffile from also writing JSON
            # metadata if not explicitly requested
            # (https://github.com/cgohlke/tifffile/issues/21)
            kwds["metadata"] = None

    if "General" in metadata.keys() and metadata["General"].get("date"):
        dt = get_date_time_from_metadata(signal["metadata"], formatting="datetime")
        kwds["datetime"] = dt

    imwrite(filename, data, software="hyperspy", photometric=photometric, **kwds)


file_writer.__doc__ %= (FILENAME_DOC.replace("read", "write to"), SIGNAL_DOC)


def file_reader(
    filename,
    lazy=False,
    force_read_resolution=False,
    multipage_as_list=False,
    hamamatsu_streak_axis_type=None,
    **kwds,
):
    """
    Read data from tif files using Christoph Gohlke's ``tifffile`` library.

    The units and the scale of images saved with ImageJ or Digital
    Micrograph is read. There is limited support for reading the scale of
    files created with Zeiss and FEI SEMs.

    Parameters
    ----------
    %s
    %s
    force_read_resolution : bool, default=False
        Force read image resolution using the ``x_resolution``, ``y_resolution``
        and ``resolution_unit`` tiff tags. Beware: most software does not (properly)
        use these tags when saving ``.tiff`` files.
        See `<https://www.loc.gov/preservation/digital/formats/content/tiff_tags.shtml>`_.
    multipage_as_list : bool, default=False
        Read multipage tiff images and return list with full content of every page. This
        utilises ``pages`` from ``tifffile`` instead of ``series`` for data access,
        which in contrast to ``series`` is able to return metadata per page.
        ``series`` (default) is only able to access metadata from the first page.
        This is recommended to be used when data is from dynamic experiments (where
        some parameters of the instrument are changing during acquisition).
    hamamatsu_streak_axis_type : str, default=None
        Select the type of time axis for hamamatsu streak files:

        - ``"uniform"``: the best-fit linear axis is used, inducing a (small)
          linearisation error. Initialise a ``UniformDataAxis``.
        - ``"data"``: the raw time axis parsed from the metadata is used.
          Initialise a ``DataAxis``.
        - ``"functional"``: the best-fit 3rd-order polynomial axis is used,
          avoiding linearisation error. Initialise a ``FunctionalDataAxis``.

        By default (``None``), ``uniform`` is used but a warning about the linearisation
        error is issued. Explicitly passing ``hamamatsu_streak_axis_type='uniform'``
        suppresses the warning. In all cases, the original axis values are stored
        in the ``original_metadata`` of the signal object. Alternatively, a
        ``DataAxis`` can be subsequently converted to a ``UniformDataAxis`` by
        interpolation using the function ``interpolate_on_axis('uniform', axis=1)``.
    **kwds : dict, optional
        Additional arguments to be passed to the ``TiffFile`` class of the `tifffile
        library <https://github.com/cgohlke/tifffile>`_.

    %s

    Examples
    --------
    >>> # Force read image resolution using the x_resolution, y_resolution and
    >>> # the resolution_unit of the TIFF tags.
    >>> s = file_reader('file.tif', force_read_resolution=True)
    >>> # Load a non-uniform axis from a hamamatsu streak file:
    >>> s = file_reader('file.tif', hamamatsu_streak_axis_type='data')
    """
    # We can't use context manager, because it closes the file on exit
    # and the file needs to stay open when loading lazily
    # close the file manually
    tiff = TiffFile(filename, **kwds)
    if multipage_as_list:
        handles = tiff.pages  # use full access with pages interface
    else:
        handles = tiff.series  # use fast access with series interface
    dict_list = [
        _read_tiff(
            tiff,
            handle,
            filename,
            force_read_resolution,
            lazy=lazy,
            hamamatsu_streak_axis_type=hamamatsu_streak_axis_type,
            **kwds,
        )
        for handle in handles
    ]
    if not lazy:
        tiff.close()

    return dict_list


file_reader.__doc__ %= (FILENAME_DOC, LAZY_DOC, RETURNS_DOC)


def _order_axes_by_name(names: list, scales: dict, offsets: dict, units: dict):
    """order axes by names in lists"""
    scales_new = [1.0] * len(names)
    offsets_new = [0.0] * len(names)
    units_new = [None] * len(names)
    for i, name in enumerate(names):
        if name == "height":
            scales_new[i] = scales["x"]
            offsets_new[i] = offsets["x"]
            units_new[i] = units["x"]
        elif name == "width":
            scales_new[i] = scales["y"]
            offsets_new[i] = offsets["y"]
            units_new[i] = units["y"]
        elif name in ["depth", "image series", "time"]:
            scales_new[i] = scales["z"]
            offsets_new[i] = offsets["z"]
            units_new[i] = units["z"]
    return scales_new, offsets_new, units_new


def _build_axes_dictionaries(shape, names=None, scales=None, offsets=None, units=None):
    """Build axes dictionaries from a set of lists"""
    if names is None:
        names = [""] * len(shape)
    if scales is None:
        scales = [1.0] * len(shape)
    if offsets is None:
        offsets = [0.0] * len(shape)
    if units is None:
        units = [None] * len(shape)

    navigate = [True] * len(shape)
    navigate[-2:] = (False, False)

    axes = [
        {
            "size": size,
            "name": str(name),
            "scale": scale,
            "offset": offset,
            "units": unit,
            "navigate": nav,
        }
        for size, name, scale, offset, unit, nav in zip(
            shape, names, scales, offsets, units, navigate
        )
    ]
    return axes


def _read_tiff(
    tiff,
    handle,
    filename,
    force_read_resolution=False,
    lazy=False,
    memmap=None,
    RGB_as_structured_array=True,
    hamamatsu_streak_axis_type=None,
    **kwds,
):
    """handle - one of either of TiffPage type or TiffPageSeries type"""
    axes = handle.axes
    if isinstance(handle, TiffPage):
        page = handle
    else:
        page = handle.pages[0]
    shape = handle.shape
    dtype = handle.dtype

    is_rgb = page.photometric == tifffile.PHOTOMETRIC.RGB and RGB_as_structured_array
    _logger.debug("Is RGB: %s" % is_rgb)
    if is_rgb:
        axes = axes[:-1]
        names = ["R", "G", "B", "A"]
        lastshape = shape[-1]
        dtype = np.dtype({"names": names[:lastshape], "formats": [dtype] * lastshape})
        shape = shape[:-1]

    op = {tag.name: tag.value for tag in page.tags}

    names = [axes_label_codes[axis] for axis in axes]

    _logger.debug("Tiff tags list: %s" % op)
    _logger.debug("Photometric: %s" % op["PhotometricInterpretation"])
    _logger.debug("is_imagej: {}".format(page.is_imagej))

    if hamamatsu_streak_axis_type not in [None, "functional", "data", "uniform"]:
        raise ValueError(
            "The `hamamatsu_streak_axis_type` argument only admits the "
            "values `None`, `'data'`, `'functional'` and `'uniform'`."
        )

    try:
        axes = _parse_scale_unit(
            tiff,
            page,
            op,
            shape,
            force_read_resolution,
            names,
            hamamatsu_streak_axis_type,
        )
    except Exception:
        _logger.info("Scale and units could not be imported")
        axes = _build_axes_dictionaries(shape, names)

    md = {
        "General": {"original_filename": os.path.split(filename)[1]},
        "Signal": {"signal_type": ""},
    }

    if "DateTime" in op:
        dt = None
        try:
            dt = datetime.strptime(op["DateTime"], "%Y:%m:%d %H:%M:%S")
        except Exception:
            try:
                if "ImageDescription" in op:
                    # JEOL SightX.
                    _dt = op["ImageDescription"]["DateTime"]
                    md["General"]["date"] = _dt[0:10]
                    # 1 extra digit for millisec should be removed
                    md["General"]["time"] = _dt[11:26]
                    md["General"]["time_zone"] = _dt[-6:]
                    dt = None
                else:
                    dt = datetime.strptime(op["DateTime"], "%Y/%m/%d %H:%M")
            except Exception:
                _logger.info("Date/Time is invalid : " + op["DateTime"])
        if dt is not None:
            md["General"]["date"] = dt.date().isoformat()
            md["General"]["time"] = dt.time().isoformat()

    # Get the digital micrograph intensity axis
    if _is_digital_micrograph(op):
        intensity_axis = _intensity_axis_digital_micrograph(op)
    else:
        intensity_axis = {}

    if "units" in intensity_axis:
        md["Signal"]["quantity"] = intensity_axis["units"]
    if "scale" in intensity_axis and "offset" in intensity_axis:
        dic = {
            "gain_factor": intensity_axis["scale"],
            "gain_offset": intensity_axis["offset"],
        }
        md["Signal"]["Noise_properties"] = {"Variance_linear_model": dic}

    data_args = handle, is_rgb
    if lazy:
        from dask import delayed
        from dask.array import from_delayed

        memmap = "memmap"
        val = delayed(_load_data, pure=True)(*data_args, memmap=memmap, **kwds)
        dc = from_delayed(val, dtype=dtype, shape=shape)
        # TODO: maybe just pass the memmap from tiffile?
    else:
        dc = _load_data(*data_args, memmap=memmap, **kwds)

    if _is_streak_hamamatsu(op):
        op.update(
            {"ImageDescriptionParsed": _get_hamamatsu_streak_description(tiff, op)}
        )
        if importlib.util.find_spec("lumispy") is not None:
            md["Signal"]["signal_type"] = "LumiTransientSpectrum"  # pragma: no cover
            # covered by on-demand integration/extension tests

    metadata_mapping = get_metadata_mapping(page, op)
    if "SightX_Notes" in op:  # TODO move to get_jeol_sightx_mapping
        md["General"]["title"] = op["SightX_Notes"]
    return {
        "data": dc,
        "original_metadata": op,
        "axes": axes,
        "metadata": md,
        "mapping": metadata_mapping,
    }


def _load_data(handle, is_rgb, memmap=None, **kwds):
    dc = handle.asarray(out=memmap)
    _logger.debug("data shape: {0}".format(dc.shape))
    if is_rgb:
        from rsciio.utils import rgb

        dc = rgb.regular_array2rgbx(dc)

    return dc


def _axes_defaults():
    """Get default axes dictionaries, with offsets and scales"""
    axes_labels = ["x", "y", "z"]
    scales = {axis: 1.0 for axis in axes_labels}
    offsets = {axis: 0.0 for axis in axes_labels}
    units = {axis: None for axis in axes_labels}

    return scales, offsets, units


def _is_force_readable(op, force_read_resolution) -> bool:
    return force_read_resolution and "ResolutionUnit" in op and "XResolution" in op


def _axes_force_read(op, shape, names):
    scales, offsets, units = _axes_defaults()
    res_unit_tag = op["ResolutionUnit"]
    if res_unit_tag != tifffile.RESUNIT.NONE:
        _logger.debug("Resolution unit: %s" % res_unit_tag)
        scales["x"], scales["y"] = _get_scales_from_x_y_resolution(op)
        # conversion to µm:
        if res_unit_tag == tifffile.RESUNIT.INCH:
            for key in ["x", "y"]:
                units[key] = "µm"
                scales[key] = scales[key] * 25400
        elif res_unit_tag == tifffile.RESUNIT.CENTIMETER:
            for key in ["x", "y"]:
                units[key] = "µm"
                scales[key] = scales[key] * 10000

    scales, offsets, units = _order_axes_by_name(names, scales, offsets, units)

    axes = _build_axes_dictionaries(shape, names, scales, offsets, units)

    return axes


def _is_fei(tiff) -> bool:
    return "fei" in tiff.flags


def _axes_fei(tiff, op, shape, names):
    _logger.debug("Reading FEI tif metadata")

    scales, offsets, units = _axes_defaults()

    op["fei_metadata"] = tiff.fei_metadata
    try:
        del op["FEI_HELIOS"]
    except KeyError:
        del op["FEI_SFEG"]
    try:
        scales["x"] = float(op["fei_metadata"]["Scan"]["PixelWidth"])
        scales["y"] = float(op["fei_metadata"]["Scan"]["PixelHeight"])
        units.update({"x": "m", "y": "m"})
    except KeyError:
        _logger.debug(
            "No 'Scan' information found in FEI metadata; attempting to get pixel size "
            "from 'IRBeam' metadata"
        )
        try:
            scales["x"] = float(op["fei_metadata"]["IRBeam"]["HFW"]) / float(
                op["fei_metadata"]["Image"]["ResolutionX"]
            )
            scales["y"] = float(op["fei_metadata"]["IRBeam"]["VFW"]) / float(
                op["fei_metadata"]["Image"]["ResolutionY"]
            )
            units.update({"x": "m", "y": "m"})
        except KeyError:
            _logger.warning(
                "Could not determine pixel size; resulting Signal will not be calibrated"
            )

    scales, offsets, units = _order_axes_by_name(names, scales, offsets, units)

    axes = _build_axes_dictionaries(shape, names, scales, offsets, units)

    return axes


def _is_zeiss(tiff) -> bool:
    return "sem" in tiff.flags


def _axes_zeiss(tiff, op, shape, names):
    _logger.debug("Reading Zeiss tif pixel_scale")
    scales, offsets, units = _axes_defaults()
    # op['CZ_SEM'][''] is containing structure of primary
    # not described SEM parameters in SI units.
    # tifffiles returns flattened version of the structure (as tuple)
    # and the scale in it is at index 3.
    # The scale is tied with physical display and needs to be multiplied
    # with factor, which is the 1024 (1k) divide by horizontal pixel n.
    # CZ_SEM tiff can contain resolution of lesser precision
    # in the described tags as 'ap_image_pixel_size' and/or
    # 'ap_pixel_size', which depending from ZEISS software version
    # can be absent and thus is not used here.
    scale_in_m = op["CZ_SEM"][""][3] * 1024 / tiff.pages[0].shape[1]
    scales.update({"x": scale_in_m, "y": scale_in_m})
    units.update({"x": "m", "y": "m"})

    scales, offsets, units = _order_axes_by_name(names, scales, offsets, units)

    axes = _build_axes_dictionaries(shape, names, scales, offsets, units)

    return axes
    # return scales, offsets, units, intensity_axis


def _is_tvips(tiff) -> bool:
    return "tvips" in tiff.flags


def _axes_tvips(tiff, op, shape, names):
    _logger.debug("Reading TVIPS tif metadata")

    scales, offsets, units = _axes_defaults()

    if "PixelSizeX" in op["TVIPS"] and "PixelSizeY" in op["TVIPS"]:
        _logger.debug("getting TVIPS scale from PixelSizeX")
        scales["x"] = op["TVIPS"]["PixelSizeX"]
        scales["y"] = op["TVIPS"]["PixelSizeY"]
        units.update({"x": "nm", "y": "nm"})
    else:
        _logger.debug("getting TVIPS scale from XYResolution")
        scales["x"], scales["y"] = _get_scales_from_x_y_resolution(op, factor=1e-2)
        units.update({"x": "m", "y": "m"})

    scales, offsets, units = _order_axes_by_name(names, scales, offsets, units)

    axes = _build_axes_dictionaries(shape, names, scales, offsets, units)

    return axes


def _is_olympus_sis(page) -> bool:
    return page.is_sis


def _axes_olympus_sis(page, tiff, op, shape, names):
    _logger.debug("Reading Olympus SIS tif metadata")
    scales, offsets, units = _axes_defaults()

    sis_metadata = {}
    for tag_number in [33471, 33560]:
        try:
            sis_metadata = page.tags[tag_number].value
        except Exception:
            pass
    op["Olympus_SIS_metadata"] = sis_metadata
    scales["x"] = round(float(sis_metadata["pixelsizex"]), 15)
    scales["y"] = round(float(sis_metadata["pixelsizey"]), 15)
    units.update({"x": "m", "y": "m"})

    scales, offsets, units = _order_axes_by_name(names, scales, offsets, units)

    axes = _build_axes_dictionaries(shape, names, scales, offsets, units)

    return axes


def _is_jeol_sightx(op) -> bool:
    return op.get("Make", None) == "JEOL Ltd."


def _axes_jeol_sightx(tiff, op, shape, names):
    # convert xml text to dictionary of tiff op['ImageDescription']
    import xml.etree.ElementTree as ET

    from box import Box

    from rsciio.utils.xml import XmlToDict

    scales, offsets, units = _axes_defaults()
    jeol_xml = "".join(
        [line.strip(" \r\n\t\x01\x00") for line in op["ImageDescription"].split("\n")]
    )
    jeol_xml_obj = ET.fromstring(jeol_xml)
    xml_to_dict = XmlToDict()
    jeol_dict = Box(xml_to_dict.dictionarize(jeol_xml_obj))
    op["ImageDescription"] = jeol_dict["TemReporter"]
    eos = op["ImageDescription"]["Eos"]["EosMode"]
    illumi = op["ImageDescription"]["IlluminationSystem"]
    imaging = op["ImageDescription"]["ImageFormingSystem"]

    # TEM/STEM
    is_STEM = eos == "modeASID"
    mode_strs = []
    mode_strs.append("STEM" if is_STEM else "TEM")
    mode_strs.append(illumi["ImageField"][4:])  # Bright Fields?
    if is_STEM:
        mode_strs.append(imaging["ScanningImageFormingMode"][4:])
    else:
        mode_strs.append(imaging["ImageFormingMode"][4:])
    mode_strs.append(imaging["SelectorString"])  # Mag / Camera Length
    op["SightX_Notes"] = ", ".join(mode_strs)

    res_unit_tag = op["ResolutionUnit"]
    if res_unit_tag == tifffile.RESUNIT.INCH:
        scale = 0.0254  # inch/m
    else:
        scale = 0.01  # tiff scaling, cm/m
    # TEM - MAG
    if (eos == "eosTEM") and (imaging["ModeString"] == "MAG"):
        mag = float(imaging["SelectorValue"])
        scales["x"], scales["y"] = _get_scales_from_x_y_resolution(
            op, factor=scale / mag * 1e9
        )
        units = {"x": "nm", "y": "nm", "z": "nm"}
    # TEM - DIFF
    elif (eos == "eosTEM") and (imaging["ModeString"] == "DIFF"):

        def wave_len(ht):
            m_e = 9.1093837015e-31
            e_c = 1.602176634e-19
            c = 299792458.0
            h = 6.62607015e-34

            momentum = 2 * m_e * e_c * ht * (1 + e_c * ht / (2 * m_e * c**2))
            return h / np.sqrt(momentum)

        camera_len = float(imaging["SelectorValue"])
        ht = float(op["ImageDescription"]["ElectronGun"]["AccelerationVoltage"])
        if imaging["SelectorUnitString"] == "mm":  # convert to "m"
            camera_len /= 1000
        elif imaging["SelectorUnitString"] == "cm":  # convert to "m"
            camera_len /= 100
        scale /= camera_len * wave_len(ht) * 1e9  # in nm
        scales["x"], scales["y"] = _get_scales_from_x_y_resolution(op, factor=scale)
        units = {"x": "1 / nm", "y": "1 / nm", "z": None}

    scales, offsets, units = _order_axes_by_name(names, scales, offsets, units)

    axes = _build_axes_dictionaries(shape, names, scales, offsets, units)

    return axes


def _is_streak_hamamatsu(op) -> bool:
    """Determines whether a .tiff page is likely to be a hamamatsu
    streak file based on the original op content.
    """
    is_hamatif = True

    # Check that the original op has an "Artist" field with "Copyright Hamamatsu"
    if "Artist" not in op:
        is_hamatif = False
        return is_hamatif
    else:
        artist = op["Artist"]
        if not artist.startswith("Copyright Hamamatsu"):
            is_hamatif = False
            return is_hamatif

    # Check that the original op has a "Software" corresponding to "HPD-TA"
    if "Software" not in op:
        is_hamatif = False
        return is_hamatif
    else:
        software = op["Software"]
        if not software.startswith("HPD-TA"):
            is_hamatif = False

    return is_hamatif


def _get_hamamatsu_streak_description(tiff, op):
    """Extract a dictionary recursively from the ImageDescription
    Metadata field in a Hamamatsu Streak .tiff file"""

    desc = op["ImageDescription"]
    dict_meta = {}
    reader = csv.reader(desc.splitlines(), delimiter=",", quotechar='"')
    for row in reader:
        key = row[0].strip(" []")
        key_dict = {}
        for element in row[1:]:
            spl = element.split("=")
            if len(spl) == 2:
                key_dict[spl[0]] = spl[1].strip('"')
        dict_meta[key] = key_dict

    # Scaling entry
    scaling = dict_meta["Scaling"]

    # Address in file where the X axis is saved. If x axis is "Other" (no
    # calibrated spectral axis saved in file), it just loads the axis as pixels
    if scaling["ScalingXScalingFile"].startswith("Other"):
        x_scale_address = None
    else:
        x_scale_address = int(re.findall(r"\d+", scaling["ScalingXScalingFile"])[0])
    xlen = op["ImageWidth"]

    # If focus mode is used there is no Y axis
    if scaling["ScalingYScalingFile"].startswith("Focus mode"):
        y_scale_address = None
    else:
        y_scale_address = int(re.findall(r"\d+", scaling["ScalingYScalingFile"])[0])
    ylen = op["ImageLength"]

    # Accessing the file as a binary
    fh = tiff.filehandle
    # Reading the x axis
    if x_scale_address is None:
        xax = np.arange(xlen)
    else:
        fh.seek(x_scale_address, 0)
        xax = np.fromfile(fh, dtype="f", count=xlen)
    if y_scale_address is None:
        yax = np.arange(ylen)
    else:
        fh.seek(y_scale_address, 0)
        yax = np.fromfile(fh, dtype="f", count=ylen)

    dict_meta["Scaling"]["ScalingXaxis"] = xax
    dict_meta["Scaling"]["ScalingYaxis"] = yax

    return dict_meta


def _axes_hamamatsu_streak(tiff, op, shape, names, hamamatsu_streak_axis_type):
    _logger.debug("Reading Hamamatsu Streak Map tif metadata")

    if hamamatsu_streak_axis_type is None:
        hamamatsu_streak_axis_type = "uniform"
        warnings.warn(
            f"{tiff} contain a non linear axis. By default, "
            f"a linearized version is initialised, which can "
            f"induce errors. Use the `hamamatsu_streak_axis_type` keyword to load "
            f"either a parabolic functional axis using `hamamatsu_streak_axis_type='functional'`, "
            f"a data axis using `hamamatsu_streak_axis_type='data'`, or use `hamamatsu_streak_axis_type='uniform'`to "
            f"linearize the axis and make this warning disappear",
            UserWarning,
        )

    # Parsing the Metadata
    desc = _get_hamamatsu_streak_description(tiff, op)
    # Getting the raw axes
    xax = desc["Scaling"]["ScalingXaxis"]
    yax = desc["Scaling"]["ScalingYaxis"]

    # Axes are initialised as a list of empty dictionaries
    axes = [{}] * len(names)

    # The width axis is always linear
    [xsc, xof] = np.polyfit(np.arange(len(xax)), xax, 1)

    i = names.index("width")
    axes[i] = {
        "size": shape[i],
        "name": "width",
        "units": desc["Scaling"]["ScalingXUnit"],
        "scale": xsc,
        "offset": xof,
    }

    # The height axis is changing
    i = names.index("height")
    axes[i] = {"name": "height", "units": desc["Scaling"]["ScalingYUnit"]}
    if hamamatsu_streak_axis_type == "uniform":
        # Uniform axis initialisation
        [ysc, yof] = np.polyfit(np.arange(len(yax)), yax, 1)
        axes[i].update(
            {
                "scale": ysc,
                "offset": yof,
                "size": shape[i],
            }
        )
    elif hamamatsu_streak_axis_type == "data":
        # Data axis initialisation
        axes[i].update({"axis": yax})
    elif hamamatsu_streak_axis_type == "functional":
        # Functional axis initialisation
        xaxis = {"scale": 1, "offset": 0, "size": len(yax)}
        poly = np.polyfit(np.arange(len(yax)), yax, 3)
        axes[i].update(
            {
                "size": len(yax),
                "x": xaxis,
                "expression": "a*x**3+b*x**2+c*x+d",
                "a": poly[0],
                "b": poly[1],
                "c": poly[2],
                "d": poly[3],
            }
        )

    return axes


def _is_imagej(tiff) -> bool:
    return "imagej" in tiff.flags


def _add_axes_imagej(tiff, op, scales, offsets, units):
    imagej_metadata = tiff.imagej_metadata
    if "ImageJ" in imagej_metadata:
        _logger.debug("Reading ImageJ tif metadata")
        # ImageJ write the unit in the image description
        if "unit" in imagej_metadata:
            if imagej_metadata["unit"] == "micron":
                units.update({"x": "µm", "y": "µm"})
            scales["x"], scales["y"] = _get_scales_from_x_y_resolution(op)
        if "spacing" in imagej_metadata:
            scales["z"] = imagej_metadata["spacing"]
    return scales, offsets, units


def _is_digital_micrograph(op) -> bool:
    # for files containing DM metadata
    tags = [
        "65003",
        "65004",
        "65005",
        "65009",
        "65010",
        "65011",
        "65006",
        "65007",
        "65008",
        "65022",
        "65024",
        "65025",
    ]
    search_result = [tag in op for tag in tags]
    return any(search_result)


def _intensity_axis_digital_micrograph(op, intensity_axis=None):
    if intensity_axis is None:
        intensity_axis = {}
    if "65022" in op:
        intensity_axis["units"] = op["65022"]  # intensity units
    if "65024" in op:
        intensity_axis["offset"] = op["65024"]  # intensity offset
    if "65025" in op:
        intensity_axis["scale"] = op["65025"]  # intensity scale
    return intensity_axis


def _add_axes_digital_micrograph(op, scales, offsets, units):
    if "65003" in op:
        _logger.debug("Reading Gatan DigitalMicrograph tif metadata")
        units["y"] = op["65003"]  # x units
    if "65004" in op:
        units["x"] = op["65004"]  # y units
    if "65005" in op:
        units["z"] = op["65005"]  # z units
    if "65009" in op:
        scales["y"] = op["65009"]  # x scales
    if "65010" in op:
        scales["x"] = op["65010"]  # y scales
    if "65011" in op:
        scales["z"] = op["65011"]  # z scales
    if "65006" in op:
        offsets["y"] = op["65006"]  # x offset
    if "65007" in op:
        offsets["x"] = op["65007"]  # y offset
    if "65008" in op:
        offsets["z"] = op["65008"]  # z offset

    return scales, offsets, units


def _parse_scale_unit(
    tiff, page, op, shape, force_read_resolution, names, hamamatsu_streak_axis_type
):
    # Force reading always has priority
    if _is_force_readable(op, force_read_resolution):
        axes = _axes_force_read(op, shape, names)
        return axes
    # Other axes readers can change position if you need to do it
    elif _is_fei(tiff):
        axes = _axes_fei(tiff, op, shape, names)
        return axes
    elif _is_zeiss(tiff):
        axes = _axes_zeiss(tiff, op, shape, names)
        return axes
    elif _is_tvips(tiff):
        axes = _axes_tvips(tiff, op, shape, names)
        return axes
    elif _is_olympus_sis(page):
        axes = _axes_olympus_sis(page, tiff, op, shape, names)
        return axes
    elif _is_jeol_sightx(op):
        axes = _axes_jeol_sightx(tiff, op, shape, names)
        return axes
    elif _is_streak_hamamatsu(op):
        axes = _axes_hamamatsu_streak(
            tiff, op, shape, names, hamamatsu_streak_axis_type
        )
        return axes
    # Axes are otherwise set to defaults
    else:
        scales, offsets, units = _axes_defaults()
        # Axes descriptors can be additionally parsed from digital micrograph or imagej-style files
        if _is_digital_micrograph(op):
            scales, offsets, units = _add_axes_digital_micrograph(
                op, scales, offsets, units
            )
        if _is_imagej(tiff):
            scales, offsets, units = _add_axes_imagej(tiff, op, scales, offsets, units)

        scales, offsets, units = _order_axes_by_name(names, scales, offsets, units)

        axes = _build_axes_dictionaries(shape, names, scales, offsets, units)

        return axes


def _get_scales_from_x_y_resolution(op, factor=1.0):
    scales = (
        op["YResolution"][1] / op["YResolution"][0] * factor,
        op["XResolution"][1] / op["XResolution"][0] * factor,
    )
    return scales


def _get_tags_dict(signal, extratags=[], factor=int(1e8)):
    """
    Get the tags to export the scale and the unit to be used in Digital
    Micrograph and ImageJ.
    """
    axes = signal["axes"]
    nav_dim = len([ax for ax in axes if ax["navigate"]])
    scales, units, offsets = _get_scale_unit(axes, encoding=None)
    _logger.debug("{0}".format(units))
    tags_dict = _get_imagej_kwargs(scales, units, nav_dim, factor=factor)
    scales, units, offsets = _get_scale_unit(axes, encoding="latin-1")

    tags_dict["extratags"].extend(
        _get_dm_kwargs_extratag(signal, scales, units, offsets, nav_dim)
    )
    tags_dict["extratags"].extend(extratags)
    return tags_dict


def _get_imagej_kwargs(scales, units, nav_dim, factor=int(1e8)):
    resolution = (
        (factor, int(scales[-1] * factor)),
        (factor, int(scales[-2] * factor)),
    )
    if nav_dim == 1:  # For stacks
        spacing = f"{scales[0]}"
    else:
        spacing = None
    description_string = _imagej_description(unit=units[1], spacing=spacing)
    _logger.debug("Description tag: {description_string}")
    extratag = [(270, "s", 1, description_string, False)]
    return {"resolution": resolution, "extratags": extratag}


def _get_dm_kwargs_extratag(signal, scales, units, offsets, nav_dim):
    extratags = [
        (65003, "s", 3, units[-1], False),  # x unit
        (65004, "s", 3, units[-2], False),  # y unit
        (65006, "d", 1, offsets[-1], False),  # x origin
        (65007, "d", 1, offsets[-2], False),  # y origin
        (65009, "d", 1, float(scales[-1]), False),  # x scale
        (65010, "d", 1, float(scales[-2]), False),
    ]  # y scale
    #                 (65012, 's', 3, units[-1], False),  # x unit full name
    #                 (65013, 's', 3, units[-2], False)]  # y unit full name
    #                 (65015, 'i', 1, 1, False), # don't know
    #                 (65016, 'i', 1, 1, False), # don't know
    #                 (65026, 'i', 1, 1, False)] # don't know
    md = DTBox(signal["metadata"], box_dots=True)
    if "Signal.quantity" in md:
        intensity_units = md["Signal"].get("quantity", "")
        extratags.extend(
            [
                (65022, "s", 3, intensity_units, False),
                (65023, "s", 3, intensity_units, False),
            ]
        )
    dic = md.get("Signal.Noise_properties.Variance_linear_model", None)
    if dic:
        try:
            intensity_offset = dic.gain_offset
            intensity_scale = dic.gain_factor
        except Exception:
            _logger.info(
                "The scale or the offset of the 'intensity axes'"
                "couldn't be retrieved, please report the bug."
            )
            intensity_offset = 0.0
            intensity_scale = 1.0
        extratags.extend(
            [
                (65024, "d", 1, intensity_offset, False),
                (65025, "d", 1, intensity_scale, False),
            ]
        )
    if nav_dim > 0:
        extratags.extend(
            [
                (65005, "s", 3, units[0], False),  # z unit
                (65008, "d", 1, offsets[0], False),  # z origin
                (65011, "d", 1, float(scales[0]), False),  # z scale
                # (65014, 's', 3, units[0], False), # z unit full name
                (65017, "i", 1, 1, False),
            ]
        )
    return extratags


def _get_scale_unit(axes, encoding=None):
    """
    Return a list of scales and units, the length of the list is equal to
    the signal dimension.

    Parameters
    ----------
    axes : list
        List of dictionary of axes

    Returns
    -------
    scales, units, offsets : list
        List of scales, units and offsets
    """
    scales = [ax["scale"] for ax in axes]
    units = [ax["units"] for ax in axes]
    offsets = [ax["offset"] for ax in axes]
    for i, unit in enumerate(units):
        if unit is None:
            units[i] = ""
        if encoding is not None:
            units[i] = units[i].encode(encoding)
    return scales, units, offsets


def _imagej_description(version="1.11a", **kwargs):
    """Return a string that will be used by ImageJ to read the unit when
    appropriate arguments are provided."""
    result = ["ImageJ=%s" % version]

    append = []
    if kwargs["spacing"] is None:
        kwargs.pop("spacing")
    for key, value in list(kwargs.items()):
        if value == "µm":
            value = "micron"
        if value == "Å":
            value = "angstrom"
        append.append(f"{key.lower()}={value}")

    return "\n".join(result + append + [""])


def _parse_beam_current_FEI(value):
    try:
        return float(value) * 1e9
    except ValueError:
        return None


def _parse_beam_energy_FEI(value):
    try:
        return float(value) * 1e-3
    except ValueError:
        return None


def _parse_working_distance_FEI(value):
    try:
        return float(value) * 1e3
    except ValueError:
        return None


def _parse_tuple_Zeiss(tup):
    value = tup[1]
    try:
        return float(value)
    except ValueError:
        return value


def _parse_tuple_Zeiss_with_units(tup, to_units=None):
    (value, parse_units) = tup[1:]
    if to_units is not None:
        from rsciio.utils._units import _UREG

        v = value * _UREG(parse_units)
        value = float("%.6e" % v.to(to_units).magnitude)
    return value


def _parse_tvips_time(value):
    # assuming this is the time in second
    return str(timedelta(seconds=int(value)))


def _parse_tvips_date(value):
    # get a number, such as 132122901, no idea, what it is... this is not
    # an excel serial, nor an unix time...
    return None


def _parse_string(value):
    if value == "":
        return None
    return value


mapping_fei = {
    "fei_metadata.Beam.HV": (
        "Acquisition_instrument.SEM.beam_energy",
        _parse_beam_energy_FEI,
    ),
    "fei_metadata.Stage.StageX": ("Acquisition_instrument.SEM.Stage.x", None),
    "fei_metadata.Stage.StageY": ("Acquisition_instrument.SEM.Stage.y", None),
    "fei_metadata.Stage.StageZ": ("Acquisition_instrument.SEM.Stage.z", None),
    "fei_metadata.Stage.StageR": ("Acquisition_instrument.SEM.Stage.rotation", None),
    "fei_metadata.Stage.StageT": ("Acquisition_instrument.SEM.Stage.tilt", None),
    "fei_metadata.Stage.WorkingDistance": (
        "Acquisition_instrument.SEM.working_distance",
        _parse_working_distance_FEI,
    ),
    "fei_metadata.Scan.Dwelltime": ("Acquisition_instrument.SEM.dwell_time", None),
    "fei_metadata.EBeam.BeamCurrent": (
        "Acquisition_instrument.SEM.beam_current",
        _parse_beam_current_FEI,
    ),
    "fei_metadata.System.SystemType": ("Acquisition_instrument.SEM.microscope", None),
    "fei_metadata.User.Date": (
        "General.date",
        lambda x: parser.parse(x).date().isoformat(),
    ),
    "fei_metadata.User.Time": (
        "General.time",
        lambda x: parser.parse(x).time().isoformat(),
    ),
    "fei_metadata.User.User": ("General.authors", None),
}


def get_jeol_sightx_mapping(op):
    mapping = {
        "ImageDescription.ElectronGun.AccelerationVoltage": (
            "Acquisition_instrument.TEM.beam_energy",
            lambda x: float(x) * 0.001,
        ),  # keV
        "ImageDescription.ElectronGun.BeamCurrent": (
            "Acquisition_instrument.TEM.beam_current",
            lambda x: float(x) * 0.001,
        ),  # nA
        "ImageDescription.Instruments": ("Acquisition_instrument.TEM.microscope", None),
        # Gonio Stage
        # depends on sample holder
        #    ("Acquisition_instrument.TEM.Stage.rotation", None),  #deg
        "ImageDescription.GonioStage.StagePosition.TX": (
            "Acquisition_instrument.TEM.Stage.tilt_alpha",
            None,
        ),  # deg
        "ImageDescription.GonioStage.StagePosition.TY": (
            "Acquisition_instrument.TEM.Stage.tilt_beta",
            None,
        ),  # deg
        # ToDo: MX(Motor)+PX(Piezo), MY+PY should be used
        #    'ImageDescription.GonioStage.StagePosition.MX':
        #    ("Acquisition_instrument.TEM.Stage.x", lambda x: float(x)*1E-6), # mm
        #    'ImageDescription.GonioStage.StagePosition.MY':
        #    ("Acquisition_instrument.TEM.Stage.y", lambda x: float(x)*1E-6), # mm
        "ImageDescription.GonioStage.MZ": (
            "Acquisition_instrument.TEM.Stage.z",
            lambda x: float(x) * 1e-6,
        ),  # mm
        #    ("General.notes", None),
        #    ("General.title", None),
        "ImageDescription.Eos.EosMode": (
            "Acquisition_instrument.TEM.acquisition_mode",
            lambda x: "STEM" if x == "eosASID" else "TEM",
        ),
        "ImageDescription.ImageFormingSystem.SelectorValue": None,
    }
    if op["ImageDescription"]["ImageFormingSystem"]["ModeString"] == "DIFF":
        mapping["ImageDescription.ImageFormingSystem.SelectorValue"] = (
            "Acquisition_instrument.TEM.camera_length",
            None,
        )
    else:  # Mag Mode
        mapping["ImageDescription.ImageFormingSystem.SelectorValue"] = (
            "Acquisition_instrument.TEM.magnification",
            None,
        )
    return mapping


mapping_cz_sem = {
    "CZ_SEM.ap_actualkv": (
        "Acquisition_instrument.SEM.beam_energy",
        _parse_tuple_Zeiss,
    ),
    "CZ_SEM.ap_mag": ("Acquisition_instrument.SEM.magnification", _parse_tuple_Zeiss),
    "CZ_SEM.ap_stage_at_x": (
        "Acquisition_instrument.SEM.Stage.x",
        lambda tup: _parse_tuple_Zeiss_with_units(tup, to_units="mm"),
    ),
    "CZ_SEM.ap_stage_at_y": (
        "Acquisition_instrument.SEM.Stage.y",
        lambda tup: _parse_tuple_Zeiss_with_units(tup, to_units="mm"),
    ),
    "CZ_SEM.ap_stage_at_z": (
        "Acquisition_instrument.SEM.Stage.z",
        lambda tup: _parse_tuple_Zeiss_with_units(tup, to_units="mm"),
    ),
    "CZ_SEM.ap_stage_at_r": (
        "Acquisition_instrument.SEM.Stage.rotation",
        _parse_tuple_Zeiss,
    ),
    "CZ_SEM.ap_stage_at_t": (
        "Acquisition_instrument.SEM.Stage.tilt",
        _parse_tuple_Zeiss,
    ),
    "CZ_SEM.ap_wd": (
        "Acquisition_instrument.SEM.working_distance",
        lambda tup: _parse_tuple_Zeiss_with_units(tup, to_units="mm"),
    ),
    "CZ_SEM.dp_dwell_time": (
        "Acquisition_instrument.SEM.dwell_time",
        lambda tup: _parse_tuple_Zeiss_with_units(tup, to_units="s"),
    ),
    "CZ_SEM.ap_iprobe": (
        "Acquisition_instrument.SEM.beam_current",
        lambda tup: _parse_tuple_Zeiss_with_units(tup, to_units="nA"),
    ),
    "CZ_SEM.dp_detector_type": (
        "Acquisition_instrument.SEM.Detector.detector_type",
        _parse_tuple_Zeiss,
    ),
    "CZ_SEM.sv_serial_number": (
        "Acquisition_instrument.SEM.microscope",
        _parse_tuple_Zeiss,
    ),
    "CZ_SEM.ap_date": (
        "General.date",
        lambda tup: parser.parse(tup[1]).date().isoformat(),
    ),
    "CZ_SEM.ap_time": (
        "General.time",
        lambda tup: parser.parse(tup[1]).time().isoformat(),
    ),
    "CZ_SEM.sv_user_name": ("General.authors", _parse_tuple_Zeiss),
}


def get_tvips_mapping(mapped_magnification):
    mapping_tvips = {
        "TVIPS.TemMagnification": (
            "Acquisition_instrument.TEM.%s" % mapped_magnification,
            None,
        ),
        "TVIPS.CameraType": ("Acquisition_instrument.TEM.Detector.Camera.name", None),
        "TVIPS.ExposureTime": (
            "Acquisition_instrument.TEM.Detector.Camera.exposure",
            lambda x: float(x) * 1e-3,
        ),
        "TVIPS.TemHighTension": (
            "Acquisition_instrument.TEM.beam_energy",
            lambda x: float(x) * 1e-3,
        ),
        "TVIPS.Comment": ("General.notes", _parse_string),
        "TVIPS.Date": ("General.date", _parse_tvips_date),
        "TVIPS.Time": ("General.time", _parse_tvips_time),
        "TVIPS.TemStagePosition": (
            "Acquisition_instrument.TEM.Stage",
            lambda stage: {
                "x": stage[0] * 1e3,
                "y": stage[1] * 1e3,
                "z": stage[2] * 1e3,
                "tilt_alpha": stage[3],
                "tilt_beta": stage[4],
            },
        ),
    }
    return mapping_tvips


mapping_olympus_sis = {
    "Olympus_SIS_metadata.magnification": (
        "Acquisition_instrument.TEM.magnification",
        None,
    ),
    "Olympus_SIS_metadata.cameraname": (
        "Acquisition_instrument.TEM.Detector.Camera.name",
        None,
    ),
}


def get_metadata_mapping(tiff_page, op):
    if tiff_page.is_fei:
        return mapping_fei

    elif tiff_page.is_sem:
        return mapping_cz_sem

    elif tiff_page.is_tvips:
        try:
            if op["TVIPS"]["TemMode"] == 3:
                mapped_magnification = "camera_length"
            else:
                mapped_magnification = "magnification"
        except KeyError:
            mapped_magnification = "magnification"
        return get_tvips_mapping(mapped_magnification)
    elif tiff_page.is_sis:
        return mapping_olympus_sis
    elif op.get("Make", None) == "JEOL Ltd.":
        return get_jeol_sightx_mapping(op)
    else:
        return {}



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import dask.array as da
import numpy as np
import sparse

from rsciio.utils._decorator import jit_ifnumba


class DenseSliceCOO(sparse.COO):
    """Just like sparse.COO, but returning a dense array on indexing/slicing"""

    def __getitem__(self, *args, **kwargs):
        obj = super().__getitem__(*args, **kwargs)
        try:
            return obj.todense()
        except AttributeError:
            # Indexing, unlike slicing, returns directly the content
            return obj


@jit_ifnumba(cache=True)
def _stream_to_sparse_COO_array_sum_frames(
    stream_data, last_frame, shape, channels, rebin_energy=1, first_frame=0
):  # pragma: no cover
    navigation_index = 0
    frame_number = 0
    ysize, xsize = shape
    frame_size = xsize * ysize
    # workaround for empty stream, numba "doesn't support" empty list, see
    # https://github.com/numba/numba/pull/2184
    # add first element and remove it at the end
    data_list = [0]
    coords_list = [(0, 0, 0)]
    data = 0
    count_channel = None
    for value in stream_data:
        if frame_number < first_frame:
            if value != 65535:  # Same spectrum
                continue
            else:
                navigation_index += 1
                if navigation_index == frame_size:
                    frame_number += 1
                    navigation_index = 0
                continue
        # when we reach the end of the frame, reset the navigation index to 0
        if navigation_index == frame_size:
            navigation_index = 0
            frame_number += 1
            if frame_number == last_frame:
                break
        # if different of ‘65535’, add a count to the corresponding channel
        if value != 65535:  # Same spectrum
            if data:
                if value == count_channel:  # Same channel, add a count
                    data += 1
                else:  # a new channel, same spectrum—requires new coord
                    # Store previous channel
                    coords_list.append(
                        (
                            int(navigation_index // xsize),
                            int(navigation_index % xsize),
                            int(count_channel // rebin_energy),
                        )
                    )
                    data_list.append(data)
                    # Add a count to new channel
                    data = 1
                    # Update count channel as this is a new channel
                    count_channel = value

            else:  # First non-zero channel of spectrum
                data = 1
                # Update count channel as this is a new channel
                count_channel = value

        else:  # Advances one pixel
            if data:  # Only store coordinates if the spectrum was not empty
                coords_list.append(
                    (
                        int(navigation_index // xsize),
                        int(navigation_index % xsize),
                        int(count_channel // rebin_energy),
                    )
                )
                data_list.append(data)
            navigation_index += 1
            data = 0

    # Store data  at the end if any (there is no final 65535 to mark the end
    # of the stream)
    if data:  # Only store coordinates if the spectrum was not empty
        coords_list.append(
            (
                int(navigation_index // xsize),
                int(navigation_index % xsize),
                int(count_channel // rebin_energy),
            )
        )
        data_list.append(data)

    final_shape = (ysize, xsize, channels // rebin_energy)
    # Remove first element, see comments above
    coords = np.array(coords_list)[1:].T
    data = np.array(data_list)[1:]
    return coords, data, final_shape


@jit_ifnumba(cache=True)
def _stream_to_sparse_COO_array(
    stream_data, last_frame, shape, channels, rebin_energy=1, first_frame=0
):  # pragma: no cover
    navigation_index = 0
    frame_number = 0
    ysize, xsize = shape
    frame_size = xsize * ysize
    # workaround for empty stream, numba "doesn't support" empty list, see
    # https://github.com/numba/numba/pull/2184
    # add first element and remove it at the end
    data_list = [0]
    coords = [(0, 0, 0, 0)]
    data = 0
    count_channel = None
    for value in stream_data:
        if frame_number < first_frame:
            if value != 65535:  # Same spectrum
                continue
            else:
                navigation_index += 1
                if navigation_index == frame_size:
                    frame_number += 1
                    navigation_index = 0
                continue
        # when we reach the end of the frame, reset the navigation index to 0
        if navigation_index == frame_size:
            navigation_index = 0
            frame_number += 1
            if frame_number == last_frame:
                break
        # if different of ‘65535’, add a count to the corresponding channel
        if value != 65535:  # Same spectrum
            if data:
                if value == count_channel:  # Same channel, add a count
                    data += 1
                else:  # a new channel, same spectrum—requires new coord
                    # Store previous channel
                    coords.append(
                        (
                            frame_number - first_frame,
                            int(navigation_index // xsize),
                            int(navigation_index % xsize),
                            int(count_channel // rebin_energy),
                        )
                    )
                    data_list.append(data)
                    # Add a count to new channel
                    data = 1
                    # Update count channel as this is a new channel
                    count_channel = value

            else:  # First non-zero channel of spectrum
                data = 1
                # Update count channel as this is a new channel
                count_channel = value

        else:  # Advances one pixel
            if data:  # Only store coordinates if the spectrum was not empty
                coords.append(
                    (
                        frame_number - first_frame,
                        int(navigation_index // xsize),
                        int(navigation_index % xsize),
                        int(count_channel // rebin_energy),
                    )
                )
                data_list.append(data)
            navigation_index += 1
            data = 0

    # Store data at the end if any (there is no final 65535 to mark the end of
    # the stream)
    if data:  # Only store coordinates if the spectrum was not empty
        coords.append(
            (
                frame_number - first_frame,
                int(navigation_index // xsize),
                int(navigation_index % xsize),
                int(count_channel // rebin_energy),
            )
        )
        data_list.append(data)

    final_shape = (last_frame - first_frame, ysize, xsize, channels // rebin_energy)
    # Remove first element, see comments above
    coords = np.array(coords)[1:].T
    data = np.array(data_list)[1:]
    return coords, data, final_shape


def stream_to_sparse_COO_array(
    stream_data,
    spatial_shape,
    channels,
    last_frame,
    rebin_energy=1,
    sum_frames=True,
    first_frame=0,
):
    """Returns data stored in a FEI stream as a multi-dimensional COO array

    Parameters
    ----------
    stream_data: numpy array
    spatial_shape: tuple of ints
        (ysize, xsize)
    channels: ints
        Number of channels in the spectrum
    rebin_energy: int
        Rebin the spectra. The default is 1 (no rebinning applied)
    sum_frames: bool
        If True, sum all the frames

    """
    if sum_frames:
        coords, data, shape = _stream_to_sparse_COO_array_sum_frames(
            stream_data=stream_data,
            shape=spatial_shape,
            channels=channels,
            rebin_energy=rebin_energy,
            first_frame=first_frame,
            last_frame=last_frame,
        )
    else:
        coords, data, shape = _stream_to_sparse_COO_array(
            stream_data=stream_data,
            shape=spatial_shape,
            channels=channels,
            rebin_energy=rebin_energy,
            first_frame=first_frame,
            last_frame=last_frame,
        )
    dense_sparse = DenseSliceCOO(coords=coords, data=data, shape=shape)
    dask_sparse = da.from_array(dense_sparse, chunks="auto")
    return dask_sparse


@jit_ifnumba(cache=True)
def _fill_array_with_stream_sum_frames(
    spectrum_image, stream, first_frame, last_frame, rebin_energy=1
):  # pragma: no cover
    # jit speeds up this function by a factor of ~ 30
    navigation_index = 0
    frame_number = 0
    shape = spectrum_image.shape
    for count_channel in np.nditer(stream):
        # when we reach the end of the frame, reset the navigation index to 0
        if navigation_index == (shape[0] * shape[1]):
            navigation_index = 0
            frame_number += 1
            # break the for loop when we reach the last frame we want to read
            if frame_number == last_frame:
                break
        # if different of ‘65535’, add a count to the corresponding channel
        if count_channel != 65535:
            if first_frame <= frame_number:
                spectrum_image[
                    navigation_index // shape[1],
                    navigation_index % shape[1],
                    count_channel // rebin_energy,
                ] += 1
        else:
            navigation_index += 1


@jit_ifnumba(cache=True)
def _fill_array_with_stream(
    spectrum_image, stream, first_frame, last_frame, rebin_energy=1
):  # pragma: no cover
    navigation_index = 0
    frame_number = 0
    shape = spectrum_image.shape
    for count_channel in np.nditer(stream):
        # when we reach the end of the frame, reset the navigation index to 0
        if navigation_index == (shape[1] * shape[2]):
            navigation_index = 0
            frame_number += 1
            # break the for loop when we reach the last frame we want to read
            if frame_number == last_frame:
                break
        # if different of ‘65535’, add a count to the corresponding channel
        if count_channel != 65535:
            if first_frame <= frame_number:
                spectrum_image[
                    frame_number - first_frame,
                    navigation_index // shape[2],
                    navigation_index % shape[2],
                    count_channel // rebin_energy,
                ] += 1
        else:
            navigation_index += 1


def stream_to_array(
    stream,
    spatial_shape,
    channels,
    last_frame,
    first_frame=0,
    rebin_energy=1,
    sum_frames=True,
    dtype="uint16",
    spectrum_image=None,
):
    """Returns data stored in a FEI stream as a multi-dimensional COO array

    Parameters
    ----------
    stream: numpy array
    spatial_shape: tuple of ints
        (ysize, xsize)
    channels: ints
        Number of channels in the spectrum
    rebin_energy: int
        Rebin the spectra. The default is 1 (no rebinning applied)
    sum_frames: bool
        If True, sum all the frames
    dtype: numpy dtype
        dtype of the array where to store the data
    number_of_frame: int or None
    spectrum_image: numpy array or None
        If not None, the array provided will be filled with the data in the
        stream.

    """

    frames = last_frame - first_frame
    if not sum_frames:
        if spectrum_image is None:
            spectrum_image = np.zeros(
                (
                    frames,
                    spatial_shape[0],
                    spatial_shape[1],
                    int(channels / rebin_energy),
                ),
                dtype=dtype,
            )

        _fill_array_with_stream(
            spectrum_image=spectrum_image,
            stream=stream,
            first_frame=first_frame,
            last_frame=last_frame,
            rebin_energy=rebin_energy,
        )
    else:
        if spectrum_image is None:
            spectrum_image = np.zeros(
                (spatial_shape[0], spatial_shape[1], int(channels / rebin_energy)),
                dtype=dtype,
            )
        _fill_array_with_stream_sum_frames(
            spectrum_image=spectrum_image,
            stream=stream,
            first_frame=first_frame,
            last_frame=last_frame,
            rebin_energy=rebin_energy,
        )
    return spectrum_image


@jit_ifnumba(cache=True)
def array_to_stream(array):  # pragma: no cover
    """Convert an array to a FEI stream

    Parameters
    ----------
    array: array

    """

    channels = array.shape[-1]
    flat_array = array.ravel()
    stream_data = []
    channel = 0
    for value in flat_array:
        for j in range(value):
            stream_data.append(channel)
        channel += 1
        if channel % channels == 0:
            channel = 0
            stream_data.append(65535)
    stream_data = stream_data[:-1]  # Remove final mark
    stream_data = np.array(stream_data)
    return stream_data



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import logging
from pathlib import Path

import h5py
from packaging.version import Version

from rsciio._docstrings import (
    CHUNKS_DOC,
    COMPRESSION_HDF5_DOC,
    COMPRESSION_HDF5_NOTES_DOC,
    FILENAME_DOC,
    LAZY_DOC,
    RETURNS_DOC,
    SHOW_PROGRESSBAR_DOC,
    SIGNAL_DOC,
)
from rsciio._hierarchical import HierarchicalReader, HierarchicalWriter, version
from rsciio.utils import file
from rsciio.utils._array import is_dask_array
from rsciio.utils._context_manager import get_progress_bar_context_manager

_logger = logging.getLogger(__name__)

not_valid_format = "The file is not a valid HyperSpy hdf5 file"

current_file_version = None  # Format version of the file being read
default_version = Version(version)


class HyperspyReader(HierarchicalReader):
    _file_type = "hspy"
    _is_hdf5 = True

    def __init__(self, file):
        super().__init__(file)
        self.Dataset = h5py.Dataset
        self.Group = h5py.Group


class HyperspyWriter(HierarchicalWriter):
    """
    An object used to simplify and organize the process for
    writing a hyperspy signal.  (.hspy format)
    """

    target_size = 1e6
    _unicode_kwds = {"dtype": h5py.string_dtype()}
    _is_hdf5 = True

    def __init__(self, file, signal, expg, **kwds):
        super().__init__(file, signal, expg, **kwds)
        self.Dataset = h5py.Dataset
        self.Group = h5py.Group

    @staticmethod
    def _store_data(data, dset, group, key, chunks, show_progressbar=True):
        # Tuple of dask arrays can also be passed, in which case the task graphs
        # are merged and the data is written in a single `dask.array.store` call.
        # This is useful when saving a ragged array, where we need to write
        # the data and the shape at the same time as the ragged array must have
        # only one dimension.
        if isinstance(data, tuple):
            data = list(data)
        elif not isinstance(data, list):
            data = [
                data,
            ]
            dset = [
                dset,
            ]

        for i, (data_, dset_) in enumerate(zip(data, dset)):
            if is_dask_array(data_):
                if data_.chunks != dset_.chunks:
                    data[i] = data_.rechunk(dset_.chunks)
                if data_.ndim == 1 and data_.dtype == object:
                    # https://github.com/hyperspy/rosettasciio/issues/198
                    raise ValueError(
                        "Saving a 1-D ragged dask array to hspy is not supported yet. "
                        "Please use the .zspy extension."
                    )
                # for performance reason, we write the data later, with all data
                # at the same time in a single `dask.array.store` call
            # "write_direct" doesn't play well with empty array
            elif data_.flags.c_contiguous and data_.shape != (0,):
                dset_.write_direct(data_)
            else:
                dset_[:] = data_
        if is_dask_array(data[0]):
            import dask.array as da

            with get_progress_bar_context_manager(show_progressbar)():
                # dask.array.store of tuple helps to merge task graphs and avoid computing twice
                da.store(data, dset)

    @staticmethod
    def _get_object_dset(group, data, key, chunks, dtype=None, **kwds):
        """Creates a h5py dataset object for saving ragged data"""
        if chunks is None:  # pragma: no cover
            chunks = 1

        if dtype is None:
            test_data = data[data.ndim * (0,)]
            if is_dask_array(test_data):
                test_data = test_data.compute()
            dtype = test_data.dtype

        dset = group.require_dataset(
            key, data.shape, dtype=h5py.vlen_dtype(dtype), chunks=chunks, **kwds
        )
        return dset


def file_reader(filename, lazy=False, **kwds):
    """
    Read data from hdf5-files saved with the HyperSpy hdf5-format
    specification (``.hspy``).

    Parameters
    ----------
    %s
    %s
    **kwds : dict, optional
        The keyword arguments are passed to :py:class:`h5py.File`.

    %s
    """
    try:
        # in case blosc compression is used
        # module needs to be imported to register plugin
        import hdf5plugin  # noqa: F401
    except ImportError:
        pass
    mode = kwds.pop("mode", "r")
    f = h5py.File(filename, mode=mode, **kwds)

    reader = HyperspyReader(f)
    # Use try, except, finally to close file when an error is raised
    try:
        exp_dict_list = reader.read(lazy=lazy)
    except BaseException as err:
        raise err
    finally:
        if not lazy:
            f.close()

    return exp_dict_list


file_reader.__doc__ %= (FILENAME_DOC, LAZY_DOC, RETURNS_DOC)


def file_writer(
    filename,
    signal,
    chunks=None,
    compression="gzip",
    close_file=True,
    write_dataset=True,
    show_progressbar=True,
    **kwds,
):
    """
    Write data to HyperSpy's hdf5-format (``.hspy``).

    Parameters
    ----------
    %s
    %s
    %s
    %s
    close_file : bool, default=True
        Close the file after writing.  The file should not be closed if the data
        needs to be accessed lazily after saving.
    write_dataset : bool, default=True
        If True, write the dataset, otherwise, don't write it. Useful to
        overwrite attributes (for example ``axes_manager``) only without having
        to write the whole dataset.
    %s
    **kwds
        The keyword argument are passed to the
        :external+h5py:meth:`h5py.Group.require_dataset` function.

    Notes
    -----
    %s
    """
    if not isinstance(write_dataset, bool):
        raise ValueError("`write_dataset` argument has to be a boolean.")

    if "shuffle" not in kwds:
        # Use shuffle by default to improve compression
        kwds["shuffle"] = True

    folder = signal["tmp_parameters"].get("original_folder", "")
    fname = signal["tmp_parameters"].get("original_filename", "")
    ext = signal["tmp_parameters"].get("original_extension", "")
    if ext and not ext.startswith("."):
        # add a dot if not present
        ext = f".{ext}"
    original_path = Path(folder, f"{fname}{ext}")

    f = None
    if signal["attributes"]["_lazy"] and Path(filename).absolute() == original_path:
        f = file.get_file_handle(signal["data"], warn=False)
        if f is not None and f.mode == "r":
            # when the file is read only, force to reopen it in writing mode
            raise OSError(
                "File opened in read only mode. To overwrite file "
                "with lazy signal, use `mode='a'` when loading the "
                "signal."
            )

    if f is None:
        # with "write_dataset=False", we need mode='a', otherwise the dataset
        # will be flushed with using 'w' mode
        mode = kwds.get("mode", "w" if write_dataset else "a")
        if mode != "a" and not write_dataset:
            raise ValueError("`mode='a'` is required to use `write_dataset=False`.")
        f = h5py.File(filename, mode=mode)

    f.attrs["file_format"] = "HyperSpy"
    f.attrs["file_format_version"] = version
    exps = f.require_group("Experiments")
    title = signal["metadata"]["General"]["title"]
    group_name = title if title else "__unnamed__"
    # / is a invalid character, see https://github.com/hyperspy/hyperspy/issues/942
    if "/" in group_name:
        group_name = group_name.replace("/", "-")
    expg = exps.require_group(group_name)

    writer = HyperspyWriter(
        f,
        signal,
        expg,
        chunks=chunks,
        compression=compression,
        write_dataset=write_dataset,
        show_progressbar=show_progressbar,
        **kwds,
    )
    # Use try, except, finally to close file when an error is raised
    try:
        writer.write()
    except BaseException as err:
        raise err
    finally:
        if close_file:
            f.close()


file_writer.__doc__ %= (
    FILENAME_DOC.replace("read", "write to"),
    SIGNAL_DOC,
    CHUNKS_DOC,
    COMPRESSION_HDF5_DOC,
    SHOW_PROGRESSBAR_DOC,
    COMPRESSION_HDF5_NOTES_DOC,
)


overwrite_dataset = HyperspyWriter.overwrite_dataset



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.
"""Utility functions for path handling."""

import logging
import os
from pathlib import Path

_logger = logging.getLogger(__name__)


__all__ = [
    "append2pathname",
    "ensure_directory",
    "incremental_filename",
    "overwrite",
]


def __dir__():
    return sorted(__all__)


def append2pathname(filename, to_append):
    """
    Append a string to a path name.

    Parameters
    ----------
    filename : str
        The original file name.
    to_append : str
        The string to append to the file name.

    Returns
    -------
    pathlib.Path
        The new file name with the appended string.
    """
    p = Path(filename)
    return Path(p.parent, p.stem + to_append + p.suffix)


def incremental_filename(filename, i=1):
    """
    If a file with the same file name exists, returns a new filename that
    does not exist.

    The new file name is created by appending `-n` (where `n` is an integer)
    to path name

    Parameters
    ----------
    filename : str
        The original file name.
    i : int
       The number to be appended.

    Returns
    -------
    pathlib.Path
        The new file name with the appended number.
    """
    filename = Path(filename)

    if filename.is_file():
        new_filename = append2pathname(filename, f"-{i}")
        if new_filename.is_file():
            return incremental_filename(filename, i + 1)
        else:
            return new_filename
    else:
        return filename


def ensure_directory(path):
    """
    Check if the path exists and if it does not, creates the directory.

    Parameters
    ----------
    path : str or pathlib.Path
        The path to check and create if it does not exist.
    """
    # If it's a file path, try the parent directory instead
    p = Path(path)
    p = p.parent if p.is_file() else p

    try:
        p.mkdir(parents=True, exist_ok=False)
    except FileExistsError:
        _logger.debug(f"Directory {p} already exists. Doing nothing.")


def overwrite(filename):
    """
    If file 'filename' exists, ask for overwriting and return True or False,
    else return True.

    Parameters
    ----------
    filename : str or pathlib.Path
        File to check for overwriting.

    Returns
    -------
    bool :
        Whether to overwrite file.
    """
    if Path(filename).is_file() or (
        Path(filename).is_dir() and os.path.splitext(filename)[1] == ".zspy"
    ):
        message = f"Overwrite '{filename}' (y/n)?\n"
        try:
            answer = input(message)
            answer = answer.lower()
            while (answer != "y") and (answer != "n"):
                print("Please answer y or n.")  # noqa: T201
                answer = input(message)
            if answer.lower() == "y":
                return True
            elif answer.lower() == "n":
                return False
            else:
                return True
        except Exception:
            # We are running in the IPython notebook that does not
            # support raw_input
            _logger.info(
                "Your terminal does not support raw input. "
                "Not overwriting. "
                "To overwrite the file use `overwrite=True`"
            )
            return False
    else:
        return True



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.


# The EMD format is a hdf5 standard proposed at Lawrence Berkeley
# National Lab (see https://emdatasets.com/ for more information).
# NOT to be confused with the FEI EMD format which was developed later.

import numpy as np
import pytest

hs = pytest.importorskip("hyperspy.api")
pytest.importorskip("h5py")


def test_de5_write_load_cycle(tmp_path):
    fname = tmp_path / "test_file.de5"
    data = np.arange(20).reshape(2, 10)
    s = hs.signals.Signal1D(data)
    s.save(fname)
    assert fname.is_file()

    s2 = hs.load(fname)
    np.testing.assert_allclose(s2.data, s.data)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

# The details of the format were taken from
# https://www.biochem.mpg.de/doc_tom/TOM_Release_2008/IOfun/tom_mrcread.html
# and https://ami.scripps.edu/software/mrctools/mrc_specification.php

import glob
import logging
import os
import re

import numpy as np

from rsciio._docstrings import (
    CHUNKS_READ_DOC,
    ENDIANESS_DOC,
    FILENAME_DOC,
    LAZY_DOC,
    NAVIGATION_SHAPE,
    RETURNS_DOC,
)
from rsciio.utils import file
from rsciio.utils._array import sarray2dict
from rsciio.utils._deprecated import (
    distributed_keyword_deprecation,
    endianess_keyword_deprecation,
    mmap_mode_keyword_deprecation,
)

_logger = logging.getLogger(__name__)

# Regex to parse movie files: YYYYMMDD_acq[_optionalSuffix][_optionalMovieNum]_movie.mrc
MOVIE_RE = re.compile(
    r"^(?P<timestamp>\d{8})_"
    r"(?P<acquisitionNumber>\d+)"
    r"(?:_(?P<optionalSuffix>.+?))?"
    r"(?:_(?P<movieNum>\d+))?_movie\.mrc$"
)

# Regex to parse virtual images:
# YYYYMMDD_acq[_optionalSuffix]_virtualImageNum_name_calculationType.mrc
# Example: 20251103_23389_3_CentroidAmp_centroid_amplitude.mrc
VIRTUAL_RE = re.compile(
    r"^(?P<timestamp>\d{8})_"
    r"(?P<acquisitionNumber>\d+)"
    r"(?:_(?P<optionalSuffix>.+?))?"
    r"_(?P<virtualImageNum>(?:\d+|virt\d+))_"
    r"(?P<name>[^_]+)_"
    r"(?P<calculationType>[^.]+)\.mrc$"
)


def find_related_de_files(movie_filename: str):
    """
    Given a DE movie filename, find related files:
    - info_file: `YYYYMMDD_acq[_optionalSuffix]_info.txt`
    - virtual_images: `YYYYMMDD_acq[_optionalSuffix]_virtualImageNum_name_calculationType.mrc`
                      (virtualImageNum supports '<digits>' or 'virt<digits>')
    - external_images: `YYYYMMDD_acq[_optionalSuffix]_ext<#>[_Name].mrc`
    - scan_csv: `YYYYMMDD_acq[_optionalSuffix]_scan_coordinates.csv`
    Returns a dict including a deduplicated list of virtual image names.
    """
    dir_name = os.path.dirname(movie_filename) or "."
    base_name = os.path.basename(movie_filename)

    # Parse timestamp and acquisitionNumber from the movie name
    m = MOVIE_RE.fullmatch(base_name)
    if m:
        ts = m.group("timestamp")
        acq = m.group("acquisitionNumber")
    else:
        return {
            "info_file": None,
            "virtual_images": [],
            "virtual_image_names": [],
            "external_images": [],
            "scan_csv": None,
        }
    prefix = f"{ts}_{acq}_"

    # Utilities
    def _first_or_none(paths):
        return sorted(paths)[0] if paths else None

    def _dedupe(seq):
        seen = {}
        out = []
        for x in seq:
            if x not in seen:
                seen[x] = 1
                out.append(x)
            else:
                count = seen[x]
                new_name = f"{x}_{count}"
                while new_name in seen:
                    count += 1
                    new_name = f"{x}_{count}"
                seen[x] = count + 1
                seen[new_name] = 1
                out.append(new_name)
        return out

    # Collect all candidate .mrc files sharing the same timestamp and acquisitionNumber
    all_mrc = glob.glob(os.path.join(dir_name, f"{prefix}*.mrc"))

    # Virtual images via regex (supports '<digits>' and 'virt<digits>')
    virtual_images = []
    virtual_image_names = []
    for p in all_mrc:
        name = os.path.basename(p)
        if name.endswith("_movie.mrc"):
            continue
        mv = VIRTUAL_RE.fullmatch(name)
        if mv:
            virtual_images.append(p)
            virtual_image_names.append(mv.group("name"))

    virtual_image_names = _dedupe(virtual_image_names)

    # External detector images (e.g., _ext3.mrc or _ext3_Name.mrc)
    ext_re = re.compile(
        rf"^{re.escape(ts)}_{re.escape(acq)}_(?:.+?_)?ext(?P<num>\d+)(?:_(?P<name>[^.]+))?\.mrc$"
    )
    external_images = []
    external_image_names = []
    for p in all_mrc:
        fname = os.path.basename(p)
        mext = ext_re.fullmatch(fname)
        if mext:
            external_images.append(p)
            ext_name = mext.group("name") or f"ext{mext.group('num')}"
            external_image_names.append(ext_name)
    external_image_names = _dedupe(external_image_names)
    # Info file
    info_file = _first_or_none(
        glob.glob(os.path.join(dir_name, f"{prefix}info.txt"))
        + glob.glob(os.path.join(dir_name, f"{prefix}*_info.txt"))
    )

    # Scan coordinates CSV
    scan_csv = _first_or_none(
        glob.glob(os.path.join(dir_name, f"{prefix}scan_coordinates.csv"))
        + glob.glob(os.path.join(dir_name, f"{prefix}*_scan_coordinates.csv"))
    )

    return {
        "info_file": info_file,
        "virtual_images": sorted(virtual_images),
        "virtual_image_names": virtual_image_names,
        "external_images": sorted(external_images),
        "external_image_names": external_image_names,
        "scan_csv": scan_csv,
    }


def get_std_dtype_list(endianness="<"):
    end = endianness
    dtype_list = [
        ("NX", end + "u4"),
        ("NY", end + "u4"),
        ("NZ", end + "u4"),
        ("MODE", end + "u4"),
        ("NXSTART", end + "u4"),
        ("NYSTART", end + "u4"),
        ("NZSTART", end + "u4"),
        ("MX", end + "u4"),
        ("MY", end + "u4"),
        ("MZ", end + "u4"),
        ("Xlen", end + "f4"),
        ("Ylen", end + "f4"),
        ("Zlen", end + "f4"),
        ("ALPHA", end + "f4"),
        ("BETA", end + "f4"),
        ("GAMMA", end + "f4"),
        ("MAPC", end + "u4"),
        ("MAPR", end + "u4"),
        ("MAPS", end + "u4"),
        ("AMIN", end + "f4"),
        ("AMAX", end + "f4"),
        ("AMEAN", end + "f4"),
        ("ISPG", end + "u2"),
        ("NSYMBT", end + "u2"),
        ("NEXT", end + "u4"),
        ("CREATID", end + "u2"),
        ("EXTRA", (np.void, 30)),
        ("NINT", end + "u2"),
        ("NREAL", end + "u2"),
        ("EXTRA2", (np.void, 28)),
        ("IDTYPE", end + "u2"),
        ("LENS", end + "u2"),
        ("ND1", end + "u2"),
        ("ND2", end + "u2"),
        ("VD1", end + "u2"),
        ("VD2", end + "u2"),
        ("TILTANGLES", (np.float32, 6)),
        ("XORIGIN", end + "f4"),
        ("YORIGIN", end + "f4"),
        ("ZORIGIN", end + "f4"),
        ("CMAP", (bytes, 4)),
        ("STAMP", (bytes, 4)),
        ("RMS", end + "f4"),
        ("NLABL", end + "u4"),
        ("LABELS", (bytes, 800)),
    ]

    return dtype_list


def get_fei_dtype_list(endianness="<"):
    end = endianness
    dtype_list = [
        ("a_tilt", end + "f4"),  # Alpha tilt (deg)
        ("b_tilt", end + "f4"),  # Beta tilt (deg)
        # Stage x position (Unit=m. But if value>1, unit=???m)
        ("x_stage", end + "f4"),
        # Stage y position (Unit=m. But if value>1, unit=???m)
        ("y_stage", end + "f4"),
        # Stage z position (Unit=m. But if value>1, unit=???m)
        ("z_stage", end + "f4"),
        # Signal2D shift x (Unit=m. But if value>1, unit=???m)
        ("x_shift", end + "f4"),
        # Signal2D shift y (Unit=m. But if value>1, unit=???m)
        ("y_shift", end + "f4"),
        ("defocus", end + "f4"),  # Defocus Unit=m. But if value>1, unit=???m)
        ("exp_time", end + "f4"),  # Exposure time (s)
        ("mean_int", end + "f4"),  # Mean value of image
        ("tilt_axis", end + "f4"),  # Tilt axis (deg)
        ("pixel_size", end + "f4"),  # Pixel size of image (m)
        ("magnification", end + "f4"),  # Magnification used
        # Not used (filling up to 128 bytes)
        ("empty", (np.void, 128 - 13 * 4)),
    ]
    return dtype_list


def get_data_type(mode):
    mode_to_dtype = {
        0: np.uint8,
        1: np.int16,
        2: np.float32,
        4: np.complex64,
        6: np.uint16,
        12: np.float16,
    }

    mode = int(mode[0])
    if mode in mode_to_dtype:
        return np.dtype(mode_to_dtype[mode])
    else:
        raise ValueError(f"Unrecognised mode '{mode}'.")


def read_de_metadata_file(filename, nav_shape=None, scan_pos_file=None):
    """This reads the metadata ".txt" file that is saved alongside a DE .mrc file.

    There are 3 ways in which the files are saved:


    Parameters
    ----------
    filename : str
        The filename of the metadata file.
    nav_shape : tuple
        The shape of the navigation axes.

    Returns
    -------
    all_lines : dict
        A dictionary containing all the metadata.
    shape : tuple
        The shape of the data in real space.
    """
    original_metadata = {}
    with open(filename) as metadata:
        for line in metadata.readlines():
            try:
                key, value = line.split("=")
                key = key.strip()
                value = value.strip()
                original_metadata[key] = value
            except ValueError:  # pragma: no cover
                _logger.warning(
                    f"Could not parse line: {line} in metadata file {filename} "
                    f"Each line should be in the form 'key = value'."
                )
    # -1 -> Not read from TEM Channel 0 -> TEM 1 -> STEM
    in_stem_mode = int(original_metadata.get("Instrument Project TEMorSTEM Mode", -1))
    scanning = original_metadata.get("Scan - Enable", "Disable") == "Enable"
    raster = original_metadata.get("Scan - Type", "Raster") == "Raster"

    if scanning and not raster:
        if scan_pos_file is None:
            raise ValueError("Scan position file is required for non-raster scans.")
        positions = np.loadtxt(scan_pos_file, delimiter=",", dtype=int)[:, ::-1]
        nav_shape = (positions[:, 0].max() + 1, positions[:, 1].max() + 1)
        unique_pos, inverse, counts = np.unique(
            positions, axis=0, return_counts=True, return_inverse=True
        )
        repeats = np.max(counts)
        if repeats > 1:
            # If there are repeated positions, we need to create a new array
            # with the repeated positions
            rep = np.zeros((len(positions), 1), dtype=int)
            positions = np.hstack((positions, rep))
            for i in range(len(positions)):
                positions[inverse == i, 2] = np.arange(len(inverse[inverse == i]))
            nav_shape = (
                repeats,
                nav_shape[0],
                nav_shape[1],
            )  # repeat the scan positions
    else:
        positions = None

    if in_stem_mode == -1:
        in_stem_mode = scanning
    elif in_stem_mode == 0:  # 0 -> TEM Mode
        in_stem_mode = False
    else:
        in_stem_mode = True

    has_camera_length = float(
        original_metadata.get("Instrument Project Camera Length (centimeters)", -1)
    )
    diffracting = (
        has_camera_length != -1 or in_stem_mode
    )  # Force diffracting if in STEM mode

    if (
        in_stem_mode and scanning and raster or nav_shape is not None
    ):  # nav-shape is not None for pos files
        axes_scales = np.array(
            [
                float(
                    original_metadata.get("Scan - Time (seconds)", 1)
                    + original_metadata.get("Scan - Repeat Delay (seconds)", 1)
                ),
                float(original_metadata.get("Specimen Pixel Size Y (nanometers)", 1)),
                float(original_metadata.get("Specimen Pixel Size X (nanometers)", 1)),
                1,  # Signal Axes below
                float(original_metadata.get("Diffraction Pixel Size Y", 1)),
                float(original_metadata.get("Diffraction Pixel Size X", 1)),
            ]
        )
        axes_scales[axes_scales != -1] = 1
        axes_units = ["sec", "nm", "nm", "times", "nm^-1", "nm^-1"]
        axes_names = ["time", "y", "x", "repeats", "ky", "kx"]
        if original_metadata.get("Scan - ROI Enable", "Disable") == "Enable":
            sizex = int(original_metadata.get("Scan - ROI Size X", 1))
            sizey = int(original_metadata.get("Scan - ROI Size Y", 1))
        else:
            sizex = int(original_metadata.get("Scan - Size X", 1))
            sizey = int(original_metadata.get("Scan - Size Y", 1))
        if nav_shape is not None and len(nav_shape) == 3:
            time = nav_shape[0]
            sizey = nav_shape[1]
            sizex = nav_shape[2]
        elif nav_shape is not None and len(nav_shape) == 2:
            sizex = nav_shape[1]
            sizey = nav_shape[0]
            time = 1
        elif nav_shape is not None and len(nav_shape) == 1:
            sizex = nav_shape[0]
            sizey = 1
            time = 1
        else:
            time = 1
        sizekx = int(original_metadata.get("Image Size X (pixels)", -1))
        sizeky = int(original_metadata.get("Image Size Y (pixels)", -1))
        navigate = [True, True, True, True, False, False]
        axes_shapes = [time, sizey, sizex, 1, sizeky, sizekx]
        nav_shape = tuple([shape for shape in [time, sizex, sizey, 1] if shape != 1])
    else:
        navigate = [True, False, False]

        nav_shape = None  # read from the .mrc file
        frame_sum = float(original_metadata.get("Autosave Movie Sum Count", 1))
        frame_time = float(original_metadata.get("Frames Per Second", 1))
        sec_per_frame = 1 / (frame_time * frame_sum)
        axes_shapes = [-1, -1, -1]  # get from the .mrc file
        if diffracting:
            axes_scales = np.array(
                [
                    sec_per_frame,
                    original_metadata.get("Diffraction Pixel Size Y", 1),
                    original_metadata.get("Diffraction Pixel Size X", 1),
                ]
            )
            axes_units = ["sec", "nm^-1", "nm^-1"]
            axes_names = ["time", "ky", "kx"]
        else:
            axes_scales = np.array(
                [
                    sec_per_frame,
                    original_metadata.get("Specimen Pixel Size Y (nanometers)", 1),
                    original_metadata.get("Specimen Pixel Size X (nanometers)", 1),
                ]
            )
            axes_units = ["sec", "nm", "nm"]
            axes_names = ["time", "y", "x"]
    axes = []
    ind = 0
    for i, s in enumerate(axes_shapes):
        if s != 1:
            if axes_scales[i] == -1:
                axes_scales[i] = 1
            axes.append(
                {
                    "name": axes_names[i],
                    "size": s,  # if -1, get from the .mrc file
                    "units": axes_units[i],
                    "scale": axes_scales[i],
                    "navigate": navigate[i],
                    "index_in_array": ind,
                }
            )
            ind += 1
    electron_gain = float(original_metadata.get("ADUs Per Electron Bin1x", 1))
    magnification = original_metadata.get("Instrument Project Magnification", None)
    camera_model = original_metadata.get("Camera Model", None)
    timestamp = original_metadata.get("Timestamp (seconds since Epoch)", None)
    fps = original_metadata.get("Frames Per Second", None)

    metadata = {
        "Acquisition_instrument": {
            "TEM": {
                "magnification": magnification,
                "detector": camera_model,
                "frames_per_second": fps,
            }
        },
        "Signal": {
            "Noise_properties": {"gain_factor": 1 / electron_gain},
            "quantity": "$e^-$",
        },
        "General": {"timestamp": timestamp},
    }
    return original_metadata, metadata, axes, nav_shape, positions


@endianess_keyword_deprecation
@distributed_keyword_deprecation
@mmap_mode_keyword_deprecation
def file_reader(
    filename,
    lazy=False,
    endianness="<",
    navigation_shape=None,
    chunks="auto",
    metadata_file="auto",
    virtual_images=None,
    external_images=None,
    scan_file=None,
):
    """
    File reader for the MRC format for tomographic and 4D-STEM data.

    Parameters
    ----------
    %s
    %s
    %s
    %s
    %s
    metadata_file : str
        The filename of the metadata file, if "auto" it will try to find the
        metadata file automatically. For DE movies of 4D STEM datasets this
        defines the shape and metadata.
    virtual_images : list
        A list of filenames of virtual images. For DE movies these are automatically loaded.
    external_images : list
        A list of filenames of external images (e.g. external detectors) to be loaded
        alongside the main data. For DE movies these are automatically loaded.
    scan_file : str
        The filename of the scan file. This is only used for DE movies to determine the scan order
        for custom scans.

    %s
    """
    virtual_image_names = None
    external_images_names = None

    if metadata_file == "auto":
        if "movie" in filename:
            info = find_related_de_files(filename)
            # Populate from related DE files helper
            metadata_file = info.get("info_file")
            virtual_images = info.get("virtual_images") or []
            virtual_image_names = info.get("virtual_image_names") or []
            external_images = info.get("external_images") or []
            scan_file = info.get("scan_csv")
            external_images_names = info.get("external_image_names") or []
        else:
            metadata_file = None

    if metadata_file is not None:
        # Check if the metadata file exists
        (
            de_metadata,
            metadata,
            metadata_axes,
            _navigation_shape,
            positions,
        ) = read_de_metadata_file(
            metadata_file, nav_shape=navigation_shape, scan_pos_file=scan_file
        )
        if navigation_shape is None:
            navigation_shape = _navigation_shape
        original_metadata = {"de_metadata": de_metadata}
    else:
        original_metadata = {}
        metadata = {"General": {}, "Signal": {}}
        metadata_axes = None
        positions = None
    metadata["General"]["original_filename"] = os.path.split(filename)[1]
    metadata["Signal"]["signal_type"] = ""
    metadata.setdefault("_HyperSpy", {})
    metadata["_HyperSpy"].setdefault("navigators", {})
    if virtual_images is not None and len(virtual_images) > 0:
        imgs = []
        for i, v in enumerate(virtual_images):
            signal = file_reader(v)[0]
            img = signal["data"]
            if navigation_shape is not None and navigation_shape[::-1] != img.shape:
                if np.prod(img.shape) == np.prod(navigation_shape[::-1]):
                    signal["data"] = img.reshape(navigation_shape[::-1])
                else:  # pragma: no cover
                    _logger.warning(
                        f"Virtual image {v} does not match the navigation shape {navigation_shape[::-1]}"
                    )
            imgs.append(signal)
            _logger.debug(f"Adding virtual image _sig_virtual_image_{i}")

            if virtual_image_names and i < len(virtual_image_names):
                name = f"_sig_{virtual_image_names[i]}"
            else:
                name = f"_sig_virtual_image_{i}"  # _sig_ is dropped in hyperspy. List  --> BaseSignal
            metadata["_HyperSpy"]["navigators"][name] = signal
        # checking to make sure the navigator is valid
        if (
            navigation_shape is not None
            and navigation_shape[::-1] == imgs[0]["data"].shape
        ):
            # Preserve existing navigator entries (virtual/external) and only
            # set the default navigator signal.
            metadata["_HyperSpy"]["_sig_navigator"] = imgs[0]

    if external_images is not None and len(external_images) > 0:
        imgs = []
        for e in external_images:
            imgs.append(file_reader(e)[0])
        for i, signal in enumerate(imgs):
            _logger.debug(f"Adding external image _sig_external_image_{i}")
            if external_images_names and i < len(external_images_names):
                name = f"_sig_{external_images_names[i]}"
            else:
                name = f"_sig_external_image_{i}"  # _sig_ is dropped in hyperspy. List  --> BaseSignal

            metadata["_HyperSpy"]["navigators"][name] = signal

    f = open(filename, "rb")
    std_header = np.fromfile(f, dtype=get_std_dtype_list(endianness), count=1)
    fei_header = None
    if std_header["NEXT"] / 1024 == 128:
        _logger.info(f"{filename} seems to contain an extended FEI header")
        fei_header = np.fromfile(f, dtype=get_fei_dtype_list(endianness), count=1024)
    if f.tell() == 1024 + std_header["NEXT"]:
        _logger.debug("The FEI header was correctly loaded")
    else:
        f.seek(1024 + std_header["NEXT"][0])
        fei_header = None
    NX, NY, NZ = std_header["NX"], std_header["NY"], std_header["NZ"]
    shape = (NX[0], NY[0], NZ[0])
    if navigation_shape is not None:
        shape = shape[:2] + navigation_shape

    data = file.memmap_distributed(
        filename,
        offset=f.tell(),
        shape=shape[::-1],
        dtype=get_data_type(std_header["MODE"]),
        positions=positions,
        chunks=chunks,
    )
    if any(shape[i] == 1 for i in range(len(shape))):
        # Remove the singleton dimensions
        data = np.squeeze(data)
    if not lazy:
        data = data.compute()

    original_metadata["std_header"] = sarray2dict(std_header)

    # Convert void/bytes fields to hex strings so they are JSON-serializable
    # (e.g. when saving to zspy/zarr). CMAP, STAMP, LABELS are bytes/void;
    # EXTRA and EXTRA2 are numpy.void padding fields.
    for key in ["EXTRA", "EXTRA2", "CMAP", "STAMP", "LABELS"]:
        val = original_metadata["std_header"].get(key)
        if val is None:
            continue
        if isinstance(val, (bytes, np.void)):
            original_metadata["std_header"][key] = bytes(val).hex()
        elif isinstance(val, np.ndarray) and val.dtype.kind == "V":
            original_metadata["std_header"][key] = val.tobytes().hex()
    if fei_header is not None:
        fei_dict = sarray2dict(
            fei_header,
        )
        del fei_dict["empty"]
        original_metadata["fei_header"] = fei_dict

    if fei_header is None:
        # The scale is in Angstroms, we convert it to nm
        scales = [
            (
                float((std_header["Zlen"] / std_header["MZ"])[0]) / 10
                if float(std_header["Zlen"][0]) != 0 and float(std_header["MZ"][0]) != 0
                else 1
            ),
            (
                float((std_header["Ylen"] / std_header["MY"])[0]) / 10
                if float(std_header["MY"][0]) != 0
                else 1
            ),
            (
                float((std_header["Xlen"] / std_header["MX"])[0]) / 10
                if float(std_header["MX"][0]) != 0
                else 1
            ),
        ]
        offsets = [
            float(std_header["ZORIGIN"][0]) / 10,
            float(std_header["YORIGIN"][0]) / 10,
            float(std_header["XORIGIN"][0]) / 10,
        ]

    else:
        # FEI does not use the standard header to store the scale
        # It does store the spatial scale in pixel_size, one per angle in
        # meters
        scales = [
            1,
        ] + [
            fei_header["pixel_size"][0] * 10**9,
        ] * 2
        offsets = [
            0,
        ] * 3

    if metadata_axes is None:
        units = [None, "nm", "nm"]
        names = ["z", "y", "x"]
        navigate = [True, False, False]
        nav_axis_to_add = 0
        if navigation_shape is not None:
            nav_axis_to_add = len(navigation_shape) - 1
            for i in range(nav_axis_to_add):
                units.insert(0, None)
                names.insert(0, "")
                navigate.insert(0, True)
                scales.insert(0, 1)
                offsets.insert(0, 0)

        # create the axis objects for each axis
        dim = len(data.shape)
        axes = [
            {
                "size": data.shape[i],
                "index_in_array": i,
                "name": names[i + nav_axis_to_add + 3 - dim],
                "scale": scales[i + nav_axis_to_add + 3 - dim],
                "offset": offsets[i + nav_axis_to_add + 3 - dim],
                "units": units[i + nav_axis_to_add + 3 - dim],
                "navigate": navigate[i + nav_axis_to_add + 3 - dim],
            }
            for i in range(dim)
        ]
    else:
        axes = metadata_axes
        for ax, s in zip(axes, shape[::-1]):
            ax["size"] = s  # Update the size of the axes

    dictionary = {
        "data": data,
        "axes": axes,
        "metadata": metadata,
        "original_metadata": original_metadata,
        "mapping": mapping,
    }

    return [
        dictionary,
    ]


mapping = {
    "fei_header.a_tilt": ("Acquisition_instrument.TEM.Stage.tilt_alpha", None),
    "fei_header.b_tilt": ("Acquisition_instrument.TEM.Stage.tilt_beta", None),
    "fei_header.x_stage": ("Acquisition_instrument.TEM.Stage.x", None),
    "fei_header.y_stage": ("Acquisition_instrument.TEM.Stage.y", None),
    "fei_header.z_stage": ("Acquisition_instrument.TEM.Stage.z", None),
    "fei_header.exp_time": (
        "Acquisition_instrument.TEM.Detector.Camera.exposure",
        None,
    ),
    "fei_header.magnification": ("Acquisition_instrument.TEM.magnification", None),
}


file_reader.__doc__ %= (
    FILENAME_DOC,
    LAZY_DOC,
    ENDIANESS_DOC,
    NAVIGATION_SHAPE,
    CHUNKS_READ_DOC,
    RETURNS_DOC,
)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

"""Test the FEI stream readers.

Because there is no official description of the format, these tests just tests
consistency between ``array_to_stream`` and ``stream_to*array``. In the
particular case of stream to sparse array, we use dask to compute the array
in order to mimic the usage in the FEI EMD reader.

"""

import numpy as np
import pytest

pytest.importorskip("h5py")
pytest.importorskip("sparse")

from rsciio.utils._fei_stream_readers import (  # noqa: E402
    array_to_stream,
    stream_to_array,
    stream_to_sparse_COO_array,
)


@pytest.mark.parametrize("lazy", (True, False))
def test_dense_stream(lazy):
    arr = np.random.randint(0, 65535, size=(2, 3, 4, 5)).astype("uint16")
    stream = array_to_stream(arr)
    if lazy:
        arrs = stream_to_sparse_COO_array(
            stream, spatial_shape=(3, 4), sum_frames=False, channels=5, last_frame=2
        )
        arrs = arrs.compute()
        assert (arrs == arr).all()
    else:
        arrs = stream_to_array(
            stream, spatial_shape=(3, 4), sum_frames=False, channels=5, last_frame=2
        )
        assert (arrs == arr).all()


@pytest.mark.parametrize("lazy", (True, False))
def test_empty_stream(lazy):
    arr = np.zeros((2, 3, 4, 5), dtype="uint16")
    stream = array_to_stream(arr)
    if lazy:
        arrs = stream_to_sparse_COO_array(
            stream, spatial_shape=(3, 4), sum_frames=False, channels=5, last_frame=2
        )
        arrs = arrs.compute()
        assert not arrs.any()
    else:
        arrs = stream_to_array(
            stream, spatial_shape=(3, 4), sum_frames=False, channels=5, last_frame=2
        )
        assert not arrs.any()


@pytest.mark.parametrize("lazy", (True, False))
def test_sparse_stream(lazy):
    arr = np.zeros((2, 3, 4, 5), dtype="uint16")
    arr[0, 0, 0, 0] = 1
    arr[-1, -1, -1, -1] = 2
    arr[1, 1, 3, 3] = 3
    stream = array_to_stream(arr)
    if lazy:
        arrs = stream_to_sparse_COO_array(
            stream, spatial_shape=(3, 4), sum_frames=False, channels=5, last_frame=2
        )
        arrs = arrs.compute()
        assert (arrs == arr).all()
    else:
        arrs = stream_to_array(
            stream, spatial_shape=(3, 4), sum_frames=False, channels=5, last_frame=2
        )
        assert (arrs == arr).all()



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import gc
import zipfile
from pathlib import Path

import numpy as np
import pytest

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")


def teardown_module(module):
    """
    Run a garbage collection cycle at the end of the test of this module
    to avoid any memory issue when continuing running the test suite.
    """
    gc.collect()


TESTS_FILE_PATH = Path(__file__).resolve().parent / "data" / "jeol"
TESTS_FILE_PATH2 = TESTS_FILE_PATH / "InvalidFrame"

TEST_FILES = [
    "rawdata.ASW",
    "View000_0000000.img",
    "View000_0000001.map",
    "View000_0000002.map",
    "View000_0000003.map",
    "View000_0000004.map",
    "View000_0000005.map",
    "View000_0000006.pts",
]

TEST_FILES2 = [
    "dummy2.ASW",
    "Dummy-Data_0000000.img",
    "Dummy-Data_0000001.map",
    "Dummy-Data_0000002.map",
    "Dummy-Data_0000003.map",
    "Dummy-Data_0000004.map",
    "Dummy-Data_0000005.map",
    "Dummy-Data_0000006.map",
    "Dummy-Data_0000007.pts",
    "Dummy-Data_0000008.apb",
    "Dummy-Data_0000009.map",
    "Dummy-Data_0000010.map",
    "Dummy-Data_0000011.map",
    "Dummy-Data_0000012.map",
    "Dummy-Data_0000013.map",
    "Dummy-Data_0000014.map",
    "Dummy-Data_0000015.pts",
    "Dummy-Data_0000016.apb",
    "Dummy-Data_0000017.map",
    "Dummy-Data_0000018.map",
    "Dummy-Data_0000019.map",
    "Dummy-Data_0000020.map",
    "Dummy-Data_0000021.map",
    "Dummy-Data_0000022.map",
    "Dummy-Data_0000023.pts",
    "Dummy-Data_0000024.APB",
]


def test_load_project():
    pytest.importorskip("numba")
    # test load all elements of the project rawdata.ASW
    filename = TESTS_FILE_PATH / TEST_FILES[0]
    s = hs.load(filename, reader="JEOL")
    # first file is always a 16bit image of the work area
    assert s[0].data.dtype == np.uint8
    assert s[0].data.shape == (512, 512)
    assert s[0].axes_manager.signal_dimension == 2
    assert s[0].axes_manager[0].units == "µm"
    assert s[0].axes_manager[0].name == "x"
    assert s[0].axes_manager[1].units == "µm"
    assert s[0].axes_manager[1].name == "y"
    # 1 to 16 files are a 16bit image of work area and elemental maps
    for elmap in s[:-1]:
        assert elmap.data.dtype == np.uint8
        assert elmap.data.shape == (512, 512)
        assert elmap.axes_manager.signal_dimension == 2
        assert elmap.axes_manager[0].units == "µm"
        assert elmap.axes_manager[0].name == "x"
        assert elmap.axes_manager[1].units == "µm"
        assert elmap.axes_manager[1].name == "y"
    # last file is the datacube
    assert s[-1].data.dtype == np.uint8
    assert s[-1].data.shape == (512, 512, 4096)
    assert s[-1].axes_manager.signal_dimension == 1
    assert s[-1].axes_manager.navigation_dimension == 2
    assert s[-1].axes_manager[0].units == "µm"
    assert s[-1].axes_manager[0].name == "x"
    assert s[-1].axes_manager[1].units == "µm"
    assert s[-1].axes_manager[1].name == "y"
    assert s[-1].axes_manager[2].units == "keV"
    np.testing.assert_allclose(
        s[-1].axes_manager[2].offset, -0.000789965 - 0.00999866 * 96
    )
    np.testing.assert_allclose(s[-1].axes_manager[2].scale, 0.00999866)
    assert s[-1].axes_manager[2].name == "Energy"

    # check scale (image)
    filename = TESTS_FILE_PATH / "Sample" / "00_View000" / TEST_FILES[1]
    s1 = hs.load(filename, reader="JEOL")
    np.testing.assert_allclose(s[0].axes_manager[0].scale, s1.axes_manager[0].scale)
    assert s[0].axes_manager[0].units == s1.axes_manager[0].units
    # check scale (pts)
    filename = TESTS_FILE_PATH / "Sample" / "00_View000" / TEST_FILES[7]
    s2 = hs.load(filename, reader="JEOL")
    np.testing.assert_allclose(s[6].axes_manager[0].scale, s2.axes_manager[0].scale)
    assert s[6].axes_manager[0].units == s2.axes_manager[0].units


def test_load_image():
    # test load work area haadf image
    filename = TESTS_FILE_PATH / "Sample" / "00_View000" / TEST_FILES[1]
    s = hs.load(filename, reader="JEOL")
    assert s.data.dtype == np.uint8
    assert s.data.shape == (512, 512)
    assert s.axes_manager.signal_dimension == 2
    assert s.axes_manager[0].units == "µm"
    np.testing.assert_allclose(s.axes_manager[0].scale, 0.00869140587747097)
    assert s.axes_manager[0].name == "x"
    assert s.axes_manager[1].units == "µm"
    np.testing.assert_allclose(s.axes_manager[1].scale, 0.00869140587747097)
    assert s.axes_manager[1].name == "y"


@pytest.mark.parametrize("SI_dtype", [np.int8, np.uint8])
def test_load_datacube(SI_dtype):
    pytest.importorskip("numba")
    # test load eds datacube
    filename = TESTS_FILE_PATH / "Sample" / "00_View000" / TEST_FILES[7]
    s = hs.load(filename, SI_dtype=SI_dtype, cutoff_at_kV=5, reader="JEOL")
    assert s.data.dtype == SI_dtype
    assert s.data.shape == (512, 512, 596)
    assert s.axes_manager.signal_dimension == 1
    assert s.axes_manager.navigation_dimension == 2
    assert s.axes_manager[0].units == "µm"
    np.testing.assert_allclose(s.axes_manager[0].scale, 0.00869140587747097)
    assert s.axes_manager[0].name == "x"
    assert s.axes_manager[1].units == "µm"
    np.testing.assert_allclose(s.axes_manager[1].scale, 0.00869140587747097)
    assert s.axes_manager[1].name == "y"
    assert s.axes_manager[2].units == "keV"
    np.testing.assert_allclose(s.axes_manager[2].offset, -0.000789965 - 0.00999866 * 96)
    np.testing.assert_allclose(s.axes_manager[2].scale, 0.00999866)
    assert s.axes_manager[2].name == "Energy"


def test_load_datacube_rebin_energy():
    pytest.importorskip("numba")
    filename = TESTS_FILE_PATH / "Sample" / "00_View000" / TEST_FILES[7]
    s = hs.load(filename, cutoff_at_kV=0.1, reader="JEOL")
    s_sum = s.sum()

    ref_data = hs.signals.Signal1D(np.array([3, 23, 77, 200, 487, 984, 1599, 2391]))
    np.testing.assert_allclose(s_sum.data[88:96], ref_data.data)

    rebin_energy = 8
    s2 = hs.load(filename, rebin_energy=rebin_energy, reader="JEOL")
    s2_sum = s2.sum()

    np.testing.assert_allclose(s2_sum.data[11:12], ref_data.data.sum())

    with pytest.raises(ValueError, match="must be a divisor"):
        _ = hs.load(filename, rebin_energy=10, reader="JEOL")


def test_load_datacube_cutoff_at_kV():
    pytest.importorskip("numba")
    gc.collect()
    cutoff_at_kV = 10.0
    filename = TESTS_FILE_PATH / "Sample" / "00_View000" / TEST_FILES[7]
    s = hs.load(filename, cutoff_at_kV=None, reader="JEOL")
    s2 = hs.load(filename, cutoff_at_kV=cutoff_at_kV, reader="JEOL")

    assert s2.axes_manager[-1].size == 1096
    np.testing.assert_allclose(s2.axes_manager[2].scale, 0.00999866)
    np.testing.assert_allclose(s2.axes_manager[2].offset, -0.9606613)

    np.testing.assert_allclose(s.sum().isig[:cutoff_at_kV].data, s2.sum().data)


def test_load_datacube_downsample():
    pytest.importorskip("numba")
    downsample = 8
    filename = TESTS_FILE_PATH / TEST_FILES[0]
    s = hs.load(filename, downsample=1, reader="JEOL")[-1]
    s2 = hs.load(filename, downsample=downsample, reader="JEOL")[-1]

    s_sum = s.sum(-1).rebin(scale=(downsample, downsample))
    s2_sum = s2.sum(-1)

    assert s2.axes_manager[-1].size == 4096
    np.testing.assert_allclose(s2.axes_manager[2].scale, 0.00999866)
    np.testing.assert_allclose(s2.axes_manager[2].offset, -0.9606613)

    for axis in s2.axes_manager.navigation_axes:
        assert axis.size == 64
        np.testing.assert_allclose(axis.scale, 0.069531247)
        np.testing.assert_allclose(axis.offset, 0.0)

    np.testing.assert_allclose(s_sum.data, s2_sum.data)

    with pytest.raises(ValueError, match="must be a divisor"):
        _ = hs.load(filename, downsample=10, reader="JEOL")[-1]

    with pytest.raises(
        ValueError,
        match="`downsample` can't be an iterable of length different from 2.",
    ):
        _ = hs.load(filename, downsample=[2, 2, 2], reader="JEOL")[-1]

    downsample = [8, 16]
    s = hs.load(filename, downsample=downsample, reader="JEOL")[-1]
    assert s.axes_manager["x"].size * downsample[0] == 512
    assert s.axes_manager["y"].size * downsample[1] == 512

    with pytest.raises(ValueError, match="must be a divisor"):
        _ = hs.load(filename, downsample=[256, 100], reader="JEOL")[-1]

    with pytest.raises(ValueError, match="must be a divisor"):
        _ = hs.load(filename, downsample=[100, 256], reader="JEOL")[-1]


def test_load_datacube_frames():
    pytest.importorskip("numba")
    rebin_energy = 2048
    filename = TESTS_FILE_PATH / "Sample" / "00_View000" / TEST_FILES[7]
    s = hs.load(filename, sum_frames=True, rebin_energy=rebin_energy, reader="JEOL")
    assert s.data.shape == (512, 512, 2)
    s_frame = hs.load(
        filename, sum_frames=False, rebin_energy=rebin_energy, reader="JEOL"
    )
    assert s_frame.data.shape == (14, 512, 512, 2)
    np.testing.assert_allclose(s_frame.sum(axis="Frame").data, s.data)
    np.testing.assert_allclose(
        s_frame.sum(axis=["x", "y", "Energy"]).data,
        np.array(
            [
                22355,
                21975,
                22038,
                21904,
                21846,
                22115,
                22021,
                21917,
                22123,
                21919,
                22141,
                22024,
                22086,
                21797,
            ]
        ),
    )


@pytest.mark.parametrize("filename_as_string", [True, False])
def test_load_eds_file(filename_as_string):
    pytest.importorskip("numba")
    pytest.importorskip("exspy", reason="exspy not installed.")
    filename = TESTS_FILE_PATH / "met03.EDS"
    if filename_as_string:
        filename = str(filename)
    s = hs.load(filename, reader="JEOL")
    assert s.metadata.Signal.signal_type == "EDS_TEM"
    assert isinstance(s, hs.signals.Signal1D)
    assert s.data.shape == (2048,)
    axis = s.axes_manager[0]
    assert axis.name == "Energy"
    assert axis.size == 2048
    assert axis.offset == -0.00176612
    assert axis.scale == 0.0100004

    # delete timestamp from metadata since it's runtime dependent
    del s.metadata.General.FileIO.Number_0.timestamp

    md_dict = s.metadata.as_dictionary()
    assert md_dict["General"] == {
        "original_filename": "met03.EDS",
        "time": "14:14:51",
        "date": "2018-06-25",
        "title": "EDX",
        "FileIO": {
            "0": {
                "operation": "load",
                "hyperspy_version": hs.__version__,
                "io_plugin": "rsciio.jeol",
            }
        },
    }
    TEM_dict = md_dict["Acquisition_instrument"]["TEM"]
    assert TEM_dict == {
        "beam_energy": 200.0,
        "Detector": {
            "EDS": {
                "azimuth_angle": 90.0,
                "detector_type": "EX24075JGT",
                "elevation_angle": 22.299999237060547,
                "energy_resolution_MnKa": 138.0,
                "live_time": 30.0,
            }
        },
        "Stage": {"tilt_alpha": 0.0},
    }


def test_shift_jis_encoding():
    # See https://github.com/hyperspy/hyperspy/issues/2812
    filename = TESTS_FILE_PATH / "181019-BN.ASW"
    # make sure we can open the file
    with open(filename, "br"):
        pass
    try:
        _ = hs.load(filename, reader="JEOL")
    except FileNotFoundError:
        # we don't have the other files required to open the data
        pass


def test_number_of_frames():
    pytest.importorskip("numba")
    dir1 = TESTS_FILE_PATH / "Sample" / "00_View000"
    dir2 = TESTS_FILE_PATH / "InvalidFrame" / "Sample" / "00_Dummy-Data"

    test_list = [  # dir, file, num_frames, num_valid_frames
        [dir1, TEST_FILES[7], 14, 14],
        [dir2, TEST_FILES2[8], 1, 0],
        [dir2, TEST_FILES2[16], 2, 1],
        [dir2, TEST_FILES2[24], 1, 1],
    ]

    for item in test_list:
        dirname, filename, frames, valid = item
        fname = str(dirname / filename)

        # Count number of frames including incomplete frame
        data = hs.load(
            fname,
            sum_frames=False,
            only_valid_data=False,
            downsample=[32, 32],
            rebin_energy=512,
            SI_dtype=np.int32,
            reader="JEOL",
        )
        assert data.axes_manager["Frame"].size == frames

        # Count number of valid frames
        data = hs.load(
            fname,
            sum_frames=False,
            only_valid_data=True,
            downsample=[32, 32],
            rebin_energy=512,
            SI_dtype=np.int32,
            reader="JEOL",
        )
        assert data.axes_manager["Frame"].size == valid


def test_em_image_in_pts():
    pytest.importorskip("numba")
    dir1 = TESTS_FILE_PATH
    dir2 = TESTS_FILE_PATH / "InvalidFrame"
    dir2p = dir2 / "Sample" / "00_Dummy-Data"

    # no SEM/STEM image
    s = hs.load(
        dir1 / TEST_FILES[0],
        read_em_image=False,
        only_valid_data=False,
        cutoff_at_kV=1,
        reader="JEOL",
    )
    assert len(s) == 7

    s = hs.load(
        dir1 / TEST_FILES[0],
        read_em_image=True,
        only_valid_data=False,
        cutoff_at_kV=1,
        reader="JEOL",
    )
    assert len(s) == 7

    # with SEM/STEM image
    s = hs.load(
        dir2 / TEST_FILES2[0],
        read_em_image=False,
        only_valid_data=False,
        cutoff_at_kV=1,
        reader="JEOL",
    )
    assert len(s) == 22
    s = hs.load(
        dir2 / TEST_FILES2[0],
        read_em_image=True,
        only_valid_data=False,
        cutoff_at_kV=1,
        reader="JEOL",
    )
    assert len(s) == 25
    assert (
        s[8].metadata.General.title
        == "S(T)EM Image extracted from " + s[8].metadata.General.original_filename
    )
    assert s[8].data[38, 15] == 87
    assert s[8].data[38, 16] == 0

    # integrate SEM/STEM image along frame axis
    s = hs.load(
        dir2p / TEST_FILES2[16],
        read_em_image=True,
        only_valid_data=False,
        sum_frames=True,
        cutoff_at_kV=1,
        frame_list=[0, 0, 0, 1],
        reader="JEOL",
    )
    assert s[1].data[0, 0] == 87 * 4
    assert s[1].data[63, 63] == 87 * 3

    s = hs.load(
        dir2p / TEST_FILES2[16],
        read_em_image=True,
        only_valid_data=False,
        sum_frames=False,
        cutoff_at_kV=1,
        reader="JEOL",
    )
    s2 = hs.load(
        dir2p / TEST_FILES2[16],
        read_em_image=True,
        only_valid_data=False,
        sum_frames=True,
        cutoff_at_kV=1,
        reader="JEOL",
    )
    s1 = [s[0].data.sum(axis=0), s[1].data.sum(axis=0)]
    assert np.array_equal(s1[0], s2[0].data)
    assert np.array_equal(s1[1], s2[1].data)


def test_pts_lazy():
    pytest.importorskip("sparse")
    dir2 = TESTS_FILE_PATH / "InvalidFrame"
    dir2p = dir2 / "Sample" / "00_Dummy-Data"
    s = hs.load(
        dir2p / TEST_FILES2[16],
        read_em_image=True,
        only_valid_data=False,
        sum_frames=False,
        lazy=True,
        reader="JEOL",
    )
    s1 = [s[0].data.sum(axis=0).compute(), s[1].data.sum(axis=0).compute()]
    s2 = hs.load(
        dir2p / TEST_FILES2[16],
        read_em_image=True,
        only_valid_data=False,
        sum_frames=True,
        lazy=False,
        reader="JEOL",
    )
    assert np.array_equal(s1[0], s2[0].data)
    assert np.array_equal(s1[1], s2[1].data)


def test_pts_frame_shift():
    pytest.importorskip("sparse")
    file = TESTS_FILE_PATH2 / "Sample" / "00_Dummy-Data" / TEST_FILES2[16]

    # without frame shift
    ref = hs.load(
        file,
        read_em_image=True,
        only_valid_data=False,
        sum_frames=False,
        lazy=False,
        reader="JEOL",
    )
    #         x, y, en
    points = [[24, 23, 106], [21, 16, 106]]
    values = [3, 1]
    targets = np.asarray([[2, 3, 106], [20, 3, 100], [4, 20, 100]], dtype=np.int16)

    # check values before shift
    d0 = np.zeros(len(points), dtype=np.int16)
    d1 = np.zeros(len(points), dtype=np.int16)
    d2 = np.zeros(len(points), dtype=np.int16)
    for frame, p in enumerate(points):
        d0[frame] = ref[0].data[frame, p[1], p[0], p[2]]
        assert d0[frame] == values[frame]

    for target in targets:
        sfts = np.zeros((ref[0].axes_manager["Frame"].size, 3), dtype=np.int16)
        for frame in range(ref[0].axes_manager["Frame"].size):
            origin = points[frame]
            sfts[frame] = np.asarray(target) - np.asarray(origin)
        shifts = sfts[:, [1, 0, 2]]

        # test frame shifts for dense (normal) loading
        s0 = hs.load(
            file,
            read_em_image=True,
            only_valid_data=False,
            sum_frames=False,
            frame_shifts=shifts,
            lazy=False,
            reader="JEOL",
        )

        for frame in range(s0[0].axes_manager["Frame"].size):
            origin = points[frame]
            sfts0 = s0[0].original_metadata.jeol_pts_frame_shifts[frame]
            pos = [origin[0] + sfts0[1], origin[1] + sfts0[0], origin[2] + sfts0[2]]
            d1[frame] = s0[0].data[frame, pos[1], pos[0], pos[2]]
            assert d1[frame] == d0[frame]

        # test frame shifts for lazy loading
        s1 = hs.load(
            file,
            read_em_image=True,
            only_valid_data=False,
            sum_frames=False,
            frame_shifts=shifts,
            lazy=True,
            reader="JEOL",
        )
        dt = s1[0].data.compute()
        for frame in range(s0[0].axes_manager["Frame"].size):
            origin = points[frame]
            sfts0 = s0[0].original_metadata.jeol_pts_frame_shifts[frame]
            pos = [origin[0] + sfts0[1], origin[1] + sfts0[0], origin[2] + sfts0[2]]
            d2[frame] = dt[frame, pos[1], pos[0], pos[2]]
            assert d2[frame] == d0[frame]

    # test frame shift with default values (no energy shift)
    sfts = np.array([[1, 2], [10, 3]])
    max_sfts = sfts.max(axis=0)
    min_sfts = sfts.min(axis=0)
    fs = sfts - max_sfts
    s = hs.load(
        file, frame_shifts=sfts, sum_frames=False, only_valid_data=False, reader="JEOL"
    )
    sz = min_sfts - max_sfts + ref[0].data.shape[1:3]
    assert s.data.shape == (2, sz[0], sz[1], 4096)
    for fr, sft in enumerate(fs):
        assert np.array_equal(
            s.data[fr, 20 + sft[0] : 30 + sft[0], 20 + sft[1] : 30 + sft[1], 106],
            ref[0].data[fr, 20:30, 20:30, 106],
        )


def test_broken_files(tmp_path):
    pytest.importorskip("numba")
    TEST_BROKEN_FILES = ["test.asw", "test.pts", "test.img"]
    for _file in TEST_BROKEN_FILES:
        file = tmp_path / _file
        with open(file, "w") as fd:
            fd.write("aaaaaaaa")
        if file.suffix == ".asw":
            # in case of asw, valid data can not be obtained
            with pytest.raises(ValueError, match="Not a valid JEOL asw format"):
                _ = hs.load(file, reader="JEOL")
        else:
            # just skipping broken files
            s = hs.load(file, reader="JEOL")
            assert s == []


def test_seq_eds_files(tmp_path):
    pos0 = [0.0, 0.0, -0.000132, 0.000132]
    pos = [
        [0.0, 0.0, 0.0, 0.0],
        [2.04070450e-05, -4.77886497e-05, 1.05909980e-05, -3.87475538e-05],
        [1.91154599e-05, -3.07397260e-05, -5.45048924e-05, 5.16634051e-05],
    ]
    memo = ["", "030", "035"]
    test_file = TESTS_FILE_PATH / "jeol_seq_eds_files.zip"

    with zipfile.ZipFile(test_file, "r") as zipped:
        zipped.extractall(tmp_path)

    # test reading sequential acquired EDS spectrum
    s = hs.load(tmp_path / "1" / "1.ASW", reader="JEOL")
    # check if three subfiles are in file (img, eds, eds)
    assert len(s) == 3
    # check positional information in subfiles
    for i, p in enumerate(pos):
        sampleinfo = s[i].original_metadata["asw"]["SampleInfo"]["0"]
        viewinfo = sampleinfo["ViewInfo"]["0"]
        np.testing.assert_allclose(viewinfo["PositionMM2"], pos0)
        viewdata_asw = viewinfo["ViewData"]
        viewdata = s[i].original_metadata["asw_viewdata"]
        np.testing.assert_allclose(viewdata["PositionMM2"], p)
        np.testing.assert_allclose(
            viewdata["PositionMM2"], viewdata_asw[i]["PositionMM2"]
        )
        assert viewdata["Memo"] == memo[i]
    for s_ in s[1:3]:
        assert s_.metadata.Signal.signal_type == "EDS_TEM"
        assert isinstance(s_, hs.signals.Signal1D)

    # test with broken asw file
    fname = tmp_path / "1" / "1.ASW"
    fname2 = tmp_path / "1" / "2.ASW"
    with open(fname, "rb") as f:
        data = bytearray(f.read())

    # No ViewData
    data2 = data.copy()
    data2[0x42D] = 0x30
    with open(fname2, "wb") as f:
        f.write(data2)
    dat = hs.load(fname2, reader="JEOL")
    assert len(dat) == 0

    # No ViewInfo
    data2 = data.copy()
    data2[0x1AD] = 0x30
    with open(fname2, "wb") as f:
        f.write(data2)
    dat = hs.load(fname2, reader="JEOL")
    assert len(dat) == 0

    # No SampleInfo
    data2 = data.copy()
    data2[0x6E] = 0x30
    with open(fname2, "wb") as f:
        f.write(data2)
    dat = hs.load(fname2, reader="JEOL")
    assert len(dat) == 0

    # test read for pseudo SEM eds/img data
    sub_dir = tmp_path / "1" / "Sample" / "00_View002"
    test_files = ["View002_0000000.img", "View002_0000001.eds"]

    # rewrite AccV  200 kV to 20 kV to generate pseudo SEM data
    # .img
    with open(sub_dir / test_files[0], "rb") as f:
        data = bytearray(f.read())
        data[0x75BC] = 0xA0
        data[0x75BD] = 0x41
    with open(sub_dir / ("x" + test_files[0]), "wb") as f:
        f.write(data)
    s = hs.load(sub_dir / ("x" + test_files[0]), reader="JEOL")
    assert "SEM" in s.metadata["Acquisition_instrument"]

    # .eds
    with open(sub_dir / test_files[1], "rb") as f:
        data = bytearray(f.read())
        data[0x4B13] = 0x34
    with open(sub_dir / ("x" + test_files[1]), "wb") as f:
        f.write(data)
    s = hs.load(sub_dir / ("x" + test_files[1]), reader="JEOL")
    assert s.metadata.Signal.signal_type == "EDS_SEM"
    assert isinstance(s, hs.signals.Signal1D)
    assert "SEM" in s.metadata["Acquisition_instrument"]


def test_frame_start_index(tmp_path):
    pytest.importorskip("numba")
    file = TESTS_FILE_PATH / "Sample" / "00_View000" / TEST_FILES[7]
    frame_start_index_ref = [
        0,
        49660,
        98602,
        147633,
        196414,
        245078,
        294263,
        343283,
        392081,
        441310,
        490126,
        539395,
        588409,
        637523,
        686084,
    ]

    ref = hs.load(
        file,
        sum_frames=False,
        downsample=[32, 32],
        rebin_energy=512,
        SI_dtype=np.int32,
        reader="JEOL",
    )
    frame_start_index = ref.original_metadata.jeol_pts_frame_start_index
    assert np.array_equal(frame_start_index, frame_start_index_ref)

    s = hs.load(
        file,
        frame_list=[2, 5],
        downsample=[32, 32],
        rebin_energy=512,
        SI_dtype=np.int32,
        reader="JEOL",
    )
    frame_start_index = s.original_metadata.jeol_pts_frame_start_index
    assert np.array_equal(frame_start_index[0:6], frame_start_index_ref[0:6])
    assert np.all(frame_start_index[6:] == -1)

    s = hs.load(
        file,
        frame_list=[4, 9],
        frame_start_index=frame_start_index,
        downsample=[32, 32],
        rebin_energy=512,
        SI_dtype=np.int32,
        reader="JEOL",
    )
    frame_start_index = s.original_metadata.jeol_pts_frame_start_index
    assert np.array_equal(frame_start_index[0:10], frame_start_index_ref[0:10])
    assert np.all(frame_start_index[10:] == -1)

    s = hs.load(
        file,
        frame_list=[11, 5, 20],
        sum_frames=False,
        frame_start_index=frame_start_index,
        downsample=[32, 32],
        rebin_energy=512,
        SI_dtype=np.int32,
        reader="JEOL",
    )
    assert s.data.shape == (2, 16, 16, 8)

    # test with pseudo "SEM" data
    test_file = tmp_path / "test.pts"
    with open(file, "rb") as f:
        data = bytearray(f.read())
        # AcckV = 20 kV
        data[0x1116] = 0xA0
        data[0x1117] = 0x41
    with open(test_file, "wb") as f:
        f.write(data)
        s = hs.load(
            test_file,
            downsample=[32, 32],
            rebin_energy=512,
            SI_dtype=np.int32,
            reader="JEOL",
        )
    assert s.metadata["Signal"]["signal_type"] == "EDS_SEM"



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import gc
import importlib
from importlib.metadata import version
from pathlib import Path

import numpy as np
import pytest
from packaging.version import Version

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
t = pytest.importorskip("traits.api", reason="traits not installed")


DATA_DIR = Path(__file__).parent / "data" / "empad"
FILENAME_STACK_RAW = DATA_DIR / "series_x10.raw"
FILENAME_MAP_RAW = DATA_DIR / "scan_x4_y4.raw"
# Version 1.2.0 and 1.2.2 xml files are identical
# keep 1.2.2 version with non-square scan dimensions for testing
FILENAME_MAP_RAW_VERSION_1_2_2 = DATA_DIR / "scan_x32_y64.raw"


if importlib.util.find_spec("pyxem") and Version(version("pyxem")) >= Version("0.19"):
    EXPECTED_UNSPECIFIED_UNITS = "px"
else:
    EXPECTED_UNSPECIFIED_UNITS = t.Undefined


def _create_raw_data(filename, shape):
    size = np.prod(shape)
    data = np.arange(size).reshape(shape).astype("float32")
    data.tofile(filename)


def setup_module(module):
    _create_raw_data(FILENAME_STACK_RAW, (166400,))
    _create_raw_data(FILENAME_MAP_RAW, (4 * 4 * 130 * 128))
    _create_raw_data(FILENAME_MAP_RAW_VERSION_1_2_2, (64 * 32 * 130 * 128))


def teardown_module(module):
    # run garbage collection to release file on windows
    gc.collect()

    fs = [
        f
        for f in [FILENAME_STACK_RAW, FILENAME_MAP_RAW, FILENAME_MAP_RAW_VERSION_1_2_2]
        if f.exists()
    ]

    for f in fs:
        # remove file
        f.unlink()


@pytest.mark.parametrize("lazy", (False, True))
def test_read_stack(lazy):
    # xml file version 0.51 211118
    s = hs.load(DATA_DIR / "stack_images.xml", lazy=lazy, file_format="EMPAD")
    assert s.data.dtype == "float32"
    ref_data = np.arange(166400).reshape((10, 130, 128))[..., :128, :]
    np.testing.assert_allclose(s.data, ref_data.astype("float32"))
    signal_axes = s.axes_manager.signal_axes
    assert signal_axes[0].name == "width"
    assert signal_axes[1].name == "height"
    for axis in signal_axes:
        units = EXPECTED_UNSPECIFIED_UNITS
        assert axis.units == units
        assert axis.scale == 1.0
        assert axis.offset == -64
    navigation_axes = s.axes_manager.navigation_axes
    assert navigation_axes[0].name == "series_count"
    assert navigation_axes[0].units == "ms"
    assert navigation_axes[0].scale == 1.0
    assert navigation_axes[0].offset == 0.0

    assert s.metadata.General.date == "2019-06-07"
    assert s.metadata.General.time == "13:17:22.590279"
    assert s.metadata.Signal.signal_type == "electron_diffraction"


@pytest.mark.parametrize("lazy", (False, True))
def test_read_map(lazy):
    # xml file version 0.51 211118
    s = hs.load(DATA_DIR / "map4x4.xml", lazy=lazy, file_format="EMPAD")
    assert s.data.dtype == "float32"
    ref_data = np.arange(266240).reshape((4, 4, 130, 128))[..., :128, :]
    np.testing.assert_allclose(s.data, ref_data.astype("float32"))
    signal_axes = s.axes_manager.signal_axes
    assert signal_axes[0].name == "width"
    assert signal_axes[1].name == "height"
    for axis in signal_axes:
        assert axis.units == "1/nm"
        np.testing.assert_allclose(axis.scale, 0.1826537)
        np.testing.assert_allclose(axis.offset, -11.689837)
    navigation_axes = s.axes_manager.navigation_axes
    assert navigation_axes[0].name == "scan_x"
    assert navigation_axes[1].name == "scan_y"
    for axis in navigation_axes:
        assert axis.units == "µm"
        np.testing.assert_allclose(axis.scale, 1.1415856)
        np.testing.assert_allclose(axis.offset, 0.0)

    assert s.metadata.General.date == "2019-06-06"
    assert s.metadata.General.time == "13:30:00.164675"
    assert s.metadata.Signal.signal_type == "electron_diffraction"


@pytest.mark.parametrize("lazy", (False, True))
@pytest.mark.parametrize("q_calibration", [None, 0.3315])
def test_read_map_1_2_2(lazy, q_calibration):
    # xml file version 1.2.2 (2021-08-20))
    filename = DATA_DIR / "EMPAD_v1.2.2_x32_y64_size0.8.xml"
    s = hs.load(filename, lazy=lazy, q_calibration=q_calibration, file_format="EMPAD")
    assert s.data.dtype == "float32"
    ref_data = np.arange(34078720).reshape((64, 32, 130, 128))[..., :128, :]
    np.testing.assert_allclose(s.data, ref_data.astype("float32"))
    signal_axes = s.axes_manager.signal_axes
    assert signal_axes[0].name == "width"
    assert signal_axes[0].index_in_array == 3
    assert signal_axes[1].name == "height"
    assert signal_axes[1].index_in_array == 2

    for axis in signal_axes:
        if q_calibration is not None:
            assert axis.units == "1/nm"
            np.testing.assert_allclose(axis.scale, q_calibration)
            np.testing.assert_allclose(axis.offset, -64 * q_calibration)
        else:
            np.testing.assert_allclose(axis.scale, 1)
            np.testing.assert_allclose(axis.offset, -64)
            assert axis.units == EXPECTED_UNSPECIFIED_UNITS
    navigation_axes = s.axes_manager.navigation_axes
    assert navigation_axes[0].name == "scan_x"
    assert navigation_axes[0].index_in_array == 1
    assert navigation_axes[1].name == "scan_y"
    assert navigation_axes[1].index_in_array == 0
    for axis in navigation_axes:
        assert axis.units == "nm"
        np.testing.assert_allclose(axis.scale, 27.6131150)
        np.testing.assert_allclose(axis.offset, 0.0)

    assert s.metadata.General.date == "2026-03-25"
    assert s.metadata.General.time == "16:47:52.586728"
    assert s.metadata.Signal.signal_type == "electron_diffraction"


@pytest.mark.parametrize("lazy", (False, True))
def test_remove_nans(lazy):
    # xml file version 1.2.2 (2021-08-20))
    filename = DATA_DIR / "EMPAD_v1.2.2_x32_y64_size0.8.xml"
    s = hs.load(filename, lazy=lazy, remove_nans=True, file_format="EMPAD")
    assert np.isnan(s.data).sum() == 0



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

"""Creates Digital Micrograph scripts to generate the dm3 testing files"""

import numpy as np

dm3_data_types = {
    1: "<i2",  # 2 byte integer signed ("short")
    2: "<f4",  # 4 byte real (IEEE 754)
    3: "<c8",  # 8 byte complex
    5: "<c8",  # 8 byte complex (packed)
    6: "<u1",  # 1 byte integer unsigned ("byte")
    7: "<i4",  # 4 byte integer signed ("long")
    8: np.dtype([("R", "u1"), ("G", "u1"), ("B", "u1"), ("A", "u1")]),
    9: "<i1",  # byte integer signed
    10: "<u2",  # 2 byte integer unsigned
    11: "<u4",  # 4 byte integer unsigned
    12: "<f8",  # 8 byte real
    13: "<c16",  # byte complex
    14: "bool",  # 1 byte binary (ie 0 or 1)
    23: np.dtype([("R", "u1"), ("G", "u1"), ("B", "u1"), ("A", "u1")]),
}

dm4_data_types = {
    1: "<i2",  # 2 byte integer signed ("short")
    2: "<f4",  # 4 byte real (IEEE 754)
    3: "<c8",  # 8 byte complex
    5: "<c8",  # 8 byte complex (packed)
    6: "<u1",  # 1 byte integer unsigned ("byte")
    7: "<i4",  # 4 byte integer signed ("long")
    8: np.dtype([("R", "u1"), ("G", "u1"), ("B", "u1"), ("A", "u1")]),
    9: "<i1",  # byte integer signed
    10: "<u2",  # 2 byte integer unsigned
    11: "<u4",  # 4 byte integer unsigned
    12: "<f8",  # 8 byte real
    13: "<c16",  # byte complex
    14: "bool",  # 1 byte binary (ie 0 or 1)
    23: np.dtype([("R", "u1"), ("G", "u1"), ("B", "u1"), ("A", "u1")]),
    27: "complex64",  # not numpy: 8-Byte packed complex (FFT data)
    28: "complex128",  # not numpy: 16-Byte packed complex (FFT data)
}


def generate_1D_files(f, data_types, dmversion):
    subfolder = f"dm{dmversion}_1D_data"
    f.write(f'CreateDirectory(path+"{subfolder}")\n')
    for key in data_types.keys():
        f.write(
            'filename = path+"'
            f'\\\\{subfolder}\\\\test-{key}.dm{dmversion}"\n'
            f'im := NewImage("test", {key}, 2)\n'
            "im[0,1] = 1\n"
            "im[1,2] = 2\n"
            "im.SaveAsGatan(filename)\n"
        )


def generate_2D_files(f, data_types, dmversion):
    subfolder = f"dm{dmversion}_2D_data"
    f.write(f'CreateDirectory(path+"{subfolder}")\n')
    for key in data_types.keys():
        f.write(
            'filename = path+"'
            f'\\\\{subfolder}\\\\test-{key}.dm{dmversion}"\n'
            f'im := NewImage("test", {key}, 2, 2)\n'
            "im[0,0,1,1] = 1\n"
            "im[0,1,1,2] = 2\n"
            "im[1,0,2,1] = 3\n"
            "im[1,1,2,2] = 4\n"
            "im.SaveAsGatan(filename)\n"
        )


def generate_3D_files(f, data_types, dmversion):
    subfolder = f"dm{dmversion}_3D_data"
    f.write(f'CreateDirectory(path+"{subfolder}")\n')
    for key in data_types.keys():
        f.write(
            'filename = path+"'
            f'\\\\{subfolder}\\\\test-{key}.dm{dmversion}"\n'
            f'im := NewImage("test", {key}, 2, 2,2)\n'
            "im[0,0,0,1,1,1] = 1\n"
            "im[1,0,0,2,1,1] = 2\n"
            "im[0,1,0,1,2,1] = 3\n"
            "im[1,1,0,2,2,1] = 4\n"
            "im[0,0,1,1,1,2] = 5\n"
            "im[1,0,1,2,1,2] = 6\n"
            "im[0,1,1,1,2,2] = 7\n"
            "im[1,1,1,2,2,2] = 8\n"
            "im.SaveAsGatan(filename)\n"
        )


def generate_4D_files(f, data_types, dmversion):
    subfolder = f"dm{dmversion}_4D_data"
    f.write(f'CreateDirectory(path+"{subfolder}")\n')
    for key in data_types.keys():
        f.write(
            'filename = path+"'
            f'\\\\{subfolder}\\\\test-{key}.dm{dmversion}"\n'
            f'im := NewImage("test", %i, 2,2,2,2)\n'
            "im[0,0,0,0,1,1,1,1] = 1\n"
            "im[1,0,0,0,2,1,1,1] = 2\n"
            "im[0,1,0,0,1,2,1,1] = 3\n"
            "im[1,1,0,0,2,2,1,1] = 4\n"
            "im[0,0,1,0,1,1,2,1] = 5\n"
            "im[1,0,1,0,2,1,2,1] = 6\n"
            "im[0,1,1,0,1,2,2,1] = 7\n"
            "im[1,1,1,0,2,2,2,1] = 8\n"
            "im[0,0,0,1,1,1,1,2] = 9\n"
            "im[1,0,0,1,2,1,1,2] = 10\n"
            "im[0,1,0,1,1,2,1,2] = 11\n"
            "im[1,1,0,1,2,2,1,2] = 12\n"
            "im[0,0,1,1,1,1,2,2] = 13\n"
            "im[1,0,1,1,2,1,2,2] = 14\n"
            "im[0,1,1,1,1,2,2,2] = 15\n"
            "im[1,1,1,1,2,2,2,2] = 16\n"
            "im.SaveAsGatan(filename)\n" % (dmversion, key, dmversion, key)
        )


if __name__ == "__main__":
    with (
        open("generate_dm3_test_files.s", "w") as f1,
        open("generate_dm4_test_files.s", "w") as f2,
    ):
        for f in (f1, f2):
            f.write("image im\n")
            f.write("string filename, path\n")
            f.write("path = GetApplicationDirectory(2, 0)\n")
            f.write('result("\\nThe file will be saved to: " + path + "\\n")\n')
        for f, dmv, dt in zip((f1, f2), (3, 4), (dm3_data_types, dm4_data_types)):
            generate_1D_files(f, dt, dmv)
            generate_2D_files(f, dt, dmv)
            generate_3D_files(f, dt, dmv)
            # generate_4D_files(f, dt, dmv)
